# coding: utf-8
"""
References:
A compact 2-D full-wave finite-difference frequency-domain method for general guided wave structures.pdf
A Compact 2-D FDFD Method for Modeling Microstrip Structures With Nonuniform Grids and Perfectly Matched Layer.pdf
Lecture 12 -- Finite-Difference Analysis of Waveguides (CEM Lectures) (Used to reduce the matrix size)
"""
import os
from PySide6 import __file__ as filepath
dirname = os.path.dirname(filepath)
plugin_path = os.path.join(dirname, 'plugins', 'platforms')
os.environ['QT_QPA_PLATFORM_PLUGIN_PATH'] = plugin_path

import numpy as np
from numpy import matlib
import sys
import scipy.linalg
# import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
try:
    from .grid3 import CartesianGrid
except ImportError:
    from grid3 import CartesianGrid
import scipy.sparse
from copy import deepcopy
from scipy.interpolate import griddata
import scipy.constants as sabitler
import matplotlib.colorbar as mcolorbar
import matplotlib.cm as cm
import matplotlib.colors as mcolors
import matplotlib.patches as patches
import inspect

nz = mcolors.Normalize()
co = sabitler.speed_of_light
epso = sabitler.epsilon_0
mu_0 = sabitler.mu_0
np.set_printoptions(precision=4)

def getindice(cc, dizi, pml=0):
    """ Get the 1-based indice of cell in dizi in which point coordinate cc resides.
    The length of first pml cells of array dizi are ignored. But calculated
    indice contains these cells.

    Args:
        cc(float): coordinate of point
        dizi((list, np.ndarray)): array of cell dimensions
        pml(int,optional): number of pml cells at the beginning of the array. Default is 0.

    Returns:
        int: 1-based indice of the cell, cc resides in.
    """
    count = 0
    length = 0
    while (length < (cc * (1.0 - 1.0e-4))):
        try:
            length = length + dizi[count+pml]
        except:
            print("end of dizi")
            return count
        count = count + 1
    return count+pml

def getindiceordered(cc, cumsumarray, pml=0):
    """Faster alternative to getindice, but works only with array of node coordinates instead of array of cell lengths. This array can be generated by cumsum function of numpy. The coordinate of starting node (0) is not included in this array.

    Args:
        cc(float): coordinate of point
        cumsumarray(np.ndarray): cumulative sum of array of cell dimensions

    Returns:
        int: 1-based indice of the cell, cc resides in.

    Example:
        getindiceordered(5.5, np.cumsum(np.ones(10)))=6
    """
    return np.searchsorted(cumsumarray[pml:], cc* (1.0 - 1.0e-4), 'right')+1

class fdfd_2d:

    # finishedsim = QtCore.Signal(list)
    # finishedmesh = QtCore.Signal()

    def __init__(self, parent = None):
        # QtCore.QObject.__init__(self, parent)
        self.parts = []
        self.dx = []
        self.dy = []
        self.pmllayers = []
        self.plot_field_cb = None #colorbar

    def exn(self, i, j):
        """ indice of Ex(i,j) in field vector """
        if i > (self.Nx - 1) or i < 0:
            return None
        if j < 1 or j > (self.Ny - 1):
            return None
        # a = self.Nx * (self.Ny - 1)  # ExSayisi = HySayisi
        # b = self.Ny * (self.Nx - 1) # EySayisi = HxSayisi
        return i + self.Nx * (j - 1)

    def eyn(self, i, j):
        """ indice of Ey(i,j) in field vector """
        if j > (self.Ny - 1) or j < 0:
            return None
        if i < 1 or i > (self.Nx - 1):
            return None
        a = self.Nx * (self.Ny - 1)  # ExSayisi = HySayisi
        # b = self.Ny * (self.Nx - 1)  # EySayisi = HxSayisi
        return a + i - 1 + (self.Nx - 1) * j

    def hxn(self, i, j):
        """ indice of Hx(i,j) in field vector """
        if j == self.Ny:
            return hxn(i, self.Ny - 1)
        elif j == -1:
            return hxn(i, 0)
        if j > self.Ny - 1 or j < 0:
            return None
        if i < 1 or i > self.Nx - 1:
            return None
        # a = hxn.Nx * (hxn.Ny - 1)  # ExSayisi = HySayisi
        # b = hxn.Ny * (hxn.Nx - 1)  # EySayisi = HxSayisi
        return i - 1 + (self.Nx - 1) * j

    def hyn(self, i, j):
        """ indice of Hy(i,j) in field vector """
        if i == self.Nx:
            return hyn(self.Nx - 1, j)
        elif i == -1:
            return hyn(0, j)
        if i > self.Nx - 1 or i < 0:
            return None
        if j < 1 or j > self.Ny - 1:
            return None
        # a = hyn.Nx * (hyn.Ny - 1)  # ExSayisi = HySayisi
        b = self.Ny * (self.Nx - 1)  # EySayisi = HxSayisi
        return b + i + self.Nx * (j - 1)

    def fdfd_mesh(self, xres, yres):
        """Create FDFD mesh.

        Args:
            boxwidth (float): Width of metal box.
            boxheight (float): Height of metal box.
            xres (float): Resolution along x-axis.
            yres (float): Resolution along y-axis
            parts (list): List of parts

        Returns:
            tuple: A tuple containing the following elements:

                1. dx, 1-d array containing the length of grid cells along x
                2. dy, 1-d array containing the length of grid cells along y
                3. er, dielectric permittivity 2-d array
                4. erxx, average dielectric permittivity 2-d array along x
                5. eryy, average dielectric permittivity 2-d array along y
                6. erzz, average dielectric permittivity 2-d array along z
                7. metalnodesx, indices of x-oriented grid edges
                8. metalnodesy, indices of y-oriented grid edges
                9. metalnodesz, indices of z-oriented grid edges
                10. currentloops, indices of grid edges used to calculate current
                11. vlines, Field component and cell indices that are used to calculate voltage
        """
        boxwidth = self.boxwidth
        boxheight = self.boxheight
        parts = self.parts
        cg = CartesianGrid()
        kritikx = []
        kritiky = []

        kritikx.append(0.0)
        kritikx.append(boxwidth)
        kritiky.append(0.0)
        kritiky.append(boxheight)
        for part in parts:
            if part[0] == "dielectric":
                kritikx.append(part[2] - part[4]/ 2.0)
                kritikx.append(part[2] + part[4]/ 2.0)
                kritiky.append(part[3] - part[5]/ 2.0)
                kritiky.append(part[3] + part[5]/ 2.0)

            if part[0] == "metal":
                kritikx.append(part[2] - part[4]/ 2.0)
                kritikx.append(part[2] + part[4]/ 2.0)
                kritiky.append(part[3] - part[5]/ 2.0)
                kritiky.append(part[3] + part[5]/ 2.0)

        cg.findsubregions(kritikx, kritiky, [0])
        print("kritikx ", kritikx)
        print("kritiky ", kritiky)
        cg.nu_grid(xres, yres, 1.0)

        if (inspect.stack()[1][3] == "multilayer"):
            if cg.get_nof_sub(1) > len([part for part in parts if part[0] == "dielectric"]):
                for part in parts:
                    if part[0] == "metal":
                        metal_y_subregion = cg.get_subregion_index_of_point(part[3], 1)
                        break
                print("metal_y_subregion = ", metal_y_subregion)
                d1 = cg.return_subregion_params(1, metal_y_subregion)[0]
                d2 = cg.return_subregion_params(1, metal_y_subregion + 1)[1]
                cg.customgrid(d1, d2, metal_y_subregion + 1, 1.2, 0, 1, 0)
                d1 = cg.return_subregion_params(1, metal_y_subregion)[0]
                d2 = cg.return_subregion_params(1, metal_y_subregion - 1)[1]
                cg.customgrid(d1, d2, metal_y_subregion - 1, 1.2, 1, 1, 0)
                cg.customgrid(0.5 * xres, yres, 1, 1.2, 2, 0, 0)
                cg.customgrid(0.5 * xres, yres, 0, 1.2, 1, 0, 0)
                cg.customgrid(0.5 * xres, yres, 2, 1.2, 0, 0, 0)

        dx = list(cg.dx)
        dy = list(cg.dy)
        dxcumsum = np.cumsum(cg.dx)
        dycumsum = np.cumsum(cg.dy)
        print("dx= ", dx)
        print("dy= ", dy)
        self.Nx = Nx = len(dx)
        self.Ny = Ny = len(dy)
        # Hucrelerdeki maddelerin epsilonları
        er = np.matrix(np.ones((Nx, Ny)))
        metalnodesx = []
        metalnodesy = []
        metalnodesz = []
        currentloops = []
        vlines = []

        for part in parts:
            if part[0] == "dielectric":
                # xi1 = getindice(part[2] - (part[4]/ 2.0), dx)
                # xi2 = getindice(part[2] + (part[4]/ 2.0), dx)
                # yi1 = getindice(part[3] - (part[5]/ 2.0), dy)
                # yi2 = getindice(part[3] + (part[5]/ 2.0), dy)
                xi1 = getindiceordered(part[2] - (part[4]/ 2.0), dxcumsum)
                xi2 = getindiceordered(part[2] + (part[4]/ 2.0), dxcumsum)
                yi1 = getindiceordered(part[3] - (part[5]/ 2.0), dycumsum)
                yi2 = getindiceordered(part[3] + (part[5]/ 2.0), dycumsum)
                # kritikx.append(xi1)
                # kritikx.append(xi2)
                # kritiky.append(yi1)
                # kritiky.append(yi2)
                for i in range(xi1, xi2):
                    for j in range(yi1, yi2):
                        er[i, j] = part[6]

            if part[0] == "metal":
                #currentloop = []
                cloop = []
                # xi1 = getindice(part[2] - (part[4]/ 2.0), dx)
                # xi2 = getindice(part[2] + (part[4]/ 2.0), dx)
                # yi1 = getindice(part[3] - (part[5]/ 2.0), dy)
                # yi2 = getindice(part[3] + (part[5]/ 2.0), dy)
                xi1 = getindiceordered(part[2] - (part[4]/ 2.0), dxcumsum)
                xi2 = getindiceordered(part[2] + (part[4]/ 2.0), dxcumsum)
                yi1 = getindiceordered(part[3] - (part[5]/ 2.0), dycumsum)
                yi2 = getindiceordered(part[3] + (part[5]/ 2.0), dycumsum)
                # kritikx.append(xi1)
                # kritikx.append(xi2)
                # kritiky.append(yi1)
                # kritiky.append(yi2)
                # cc = getindice((part[2]), dx)
                cc = getindiceordered((part[2]), dxcumsum)
                # vline = [(eyn(cc, i), dy[i]) for i in range(yi1)]
                vlines.append(("eyn",cc, 0, yi1))
                for j in range(yi1, yi2 + 1):
                    # currentloop.append((hyn(xi2, j), -0.5 * (dy[j] + dy[j - 1])))
                    # currentloop.append((hyn(xi1 - 1, j), 0.5 * (dy[j] + dy[j - 1])))
                    if j==0:
                        cloop.append(("hyn",xi2, j, -0.5 * (dy[j])))
                        cloop.append(("hyn",xi1 - 1, j, 0.5 * (dy[j])))
                    elif j==len(dy):
                        cloop.append(("hyn",xi2, j, -0.5 * (dy[j - 1])))
                        cloop.append(("hyn",xi1 - 1, j, 0.5 * (dy[j - 1])))
                    else:
                        cloop.append(("hyn",xi2, j, -0.5 * (dy[j] + dy[j - 1])))
                        cloop.append(("hyn",xi1 - 1, j, 0.5 * (dy[j] + dy[j - 1])))
                for i in range(xi1, xi2 + 1):
                    # currentloop.append((hxn(i, yi2), 0.5 * (dx[i] + dx[i - 1])))
                    # currentloop.append((hxn(i, yi1 - 1), -0.5 * (dx[i] + dx[i - 1])))
                    if i == 0:
                        cloop.append(("hxn",i, yi2, 0.5 * (dx[i])))
                        cloop.append(("hxn",i, yi1 - 1, -0.5 * (dx[i])))
                    elif i ==len(dx):
                        cloop.append(("hxn",i, yi2, 0.5 * (dx[i - 1])))
                        cloop.append(("hxn",i, yi1 - 1, -0.5 * (dx[i - 1])))
                    else:
                        cloop.append(("hxn",i, yi2, 0.5 * (dx[i] + dx[i - 1])))
                        cloop.append(("hxn",i, yi1 - 1, -0.5 * (dx[i] + dx[i - 1])))
                    for j in range(yi1, yi2 + 1):
                        if j>0 and j<len(dy):
                            if i>0 and i<len(dx):
                                metalnodesz.append((i, j))
                        if i < (xi2):
                            if j>0 and j<len(dy):
                                if i>0 and i<len(dx):
                                    metalnodesx.append((i, j))
                        if j < (yi2):
                            if j>0 and j<len(dy):
                                if i>0 and i<len(dx):
                                    metalnodesy.append((i, j))
                # currentloops.append(currentloop)
                currentloops.append(cloop)

        # Effective permittivity valid for E-Fields
        erxx = np.matrix(np.ones((Nx, Ny + 1), float))
        eryy = np.matrix(np.ones((Nx + 1, Ny), float))
        erzz = np.matrix(np.ones((Nx + 1, Ny + 1), float))
        erxx[:Nx, 0] = er[:Nx, 0]
        erxx[:Nx, Ny] = er[:Nx, Ny - 1]
        erxx[:Nx, 1:Ny] = 0.5 * (er[:Nx, :Ny - 1] + er[:Nx, 1:Ny])

        eryy[0, :Ny] = er[0, :Ny]
        eryy[Nx, :Ny] = er[Nx - 1, :Ny]
        eryy[1:Nx, :Ny] = 0.5 * (er[:Nx - 1, :Ny] + er[1:Nx, :Ny])

        erzz[1:Nx, 1:Ny] = 0.25 * (er[1:Nx, 1:Ny] + er[1:Nx, :Ny - 1] + er[:Nx - 1, 1:Ny] + er[:Nx - 1, :Ny - 1])
        erzz[1:Nx, 0] = 0.5 * (er[1:Nx, 0] + er[:Nx - 1, 0])
        erzz[1:Nx, Ny] = 0.5 * (er[1:Nx, Ny - 1] + er[:Nx - 1, Ny - 1])
        erzz[0, 1:Ny] = 0.5 * (er[0, 1:Ny] + er[0, :Ny - 1])
        erzz[Nx, 1:Ny] = 0.5 * (er[Nx - 1, 1:Ny] + er[Nx - 1, :Ny - 1])

        erzz[0, 0] = er[0, 0]
        erzz[0, Ny] = er[0, Ny - 1]
        erzz[Nx, Ny] = er[Nx - 1, Ny - 1]
        erzz[Nx, 0] = er[Nx - 1, 0]
        self.dx = dx
        self.dy = dy
        self.er = er
        self.erxx = erxx
        self.eryy = eryy
        self.erzz = erzz
        self.metalnodesx = metalnodesx
        self.metalnodesy = metalnodesy
        self.metalnodesz = metalnodesz
        self.vlines = vlines
        self.currentloops = currentloops

        # self.finishedmesh.emit()
        # return (dx, dy, er, erxx, eryy, erzz, metalnodesx, metalnodesy, metalnodesz, currentloops, vlines)

    def addpml(self, pmllayers):
        """Add PML (perfectly matched layers) and return modified grid and permittivity arrays.

        Args:
            pmllayers (list of int): Number of PML layers x1,x2,y1,y2.
            dx (list of float): lengths of x-oriented grid cells
            dy (list of float): lengths of y-oriented grid cells
            er (2-d array of float): Dielectric permittivity array
            erxx (2-d array of float): Average dielectric permittivities at x-oriented grid edges
            eryy (2-d array of float): Average dielectric permittivities at y-oriented grid edges
            erzz (2-d array of float): Average dielectric permittivities at z-oriented grid edges
            metalnodesx (list of 2-tuples of int): indices of x-oriented grid edges
            metalnodesy (list of 2-tuples of int): indices of y-oriented grid edges
            metalnodesz (list of 2-tuples of int): indices of z-oriented grid edges
            currentloops (list): Indices of grid edges used to calculate current
            vlines (list): Field component and cell indices that are used to calculate voltage

        Returns:
            None
        """
        dx = self.dx
        dy = self.dy
        er = self.er
        erxx = self.erxx
        eryy = self.eryy
        erzz = self.erzz
        metalnodesx = self.metalnodesx
        metalnodesy = self.metalnodesy
        metalnodesz = self.metalnodesz
        vlines = self.vlines
        currentloops = self.currentloops
        self.pmllayers = pmllayers.copy()
        x1,x2,y1,y2=tuple(pmllayers)
        newdx=[dx[0]]*x1+dx+[dx[-1]]*x2
        newdy=[dy[0]]*y1+dy+[dy[-1]]*y2
        self.Nx = Nx = len(newdx)
        self.Ny = Ny = len(newdy)
        print("Nx, Ny ",Nx,Ny)

        newerxx = np.matrix(np.ones((Nx, Ny + 1), float))
        neweryy = np.matrix(np.ones((Nx + 1, Ny), float))
        newerzz = np.matrix(np.ones((Nx + 1, Ny + 1), float))
        newer = np.matrix(np.ones((Nx, Ny), float))
        newerxx[:x1,:y1]=erxx[0,0]
        neweryy[:x1,:y1]=eryy[0,0]
        newerzz[:x1,:y1]=erzz[0,0]
        newer[:x1,:y1]=er[0,0]
        newerxx[:x1,-y2:]=erxx[0,-1]
        neweryy[:x1,-y2:]=eryy[0,-1]
        newerzz[:x1,-y2:]=erzz[0,-1]
        newer[:x1,-y2:]=er[0,-1]
        newerxx[-x2:,:y1]=erxx[-1,0]
        neweryy[-x2:,:y1]=eryy[-1,0]
        newerzz[-x2:,:y1]=erzz[-1,0]
        newer[-x2:,:y1]=er[-1,0]
        newerxx[-x2:,-y2:]=erxx[-1,-1]
        neweryy[-x2:,-y2:]=eryy[-1,-1]
        newerzz[-x2:,-y2:]=erzz[-1,-1]
        newer[-x2:,-y2:]=er[-1,-1]
        for i in range(x1,Nx-x2):
            newerxx[i, -y2:] = erxx[i-x1, -1]
            neweryy[i, -y2:] = eryy[i-x1, -1]
            newerzz[i, -y2:] = erzz[i-x1, -1]
            newer[i, -y2:] = er[i-x1, -1]
            newerxx[i, :y1] = erxx[i-x1, 0]
            neweryy[i, :y1] = eryy[i-x1, 0]
            newerzz[i, :y1] = erzz[i-x1, 0]
            newer[i, :y1] = er[i-x1, 0]
        for j in range(y1,Ny-y2):
            newerxx[-x2:, j] = erxx[-1, j-y1]
            neweryy[-(x2+1):, j] = eryy[-1, j-y1]
            newerzz[-(x2+1):, j] = erzz[-1, j-y1]
            newer[-x2:, j] = er[-1, j-y1]
            newerxx[:x1, j] = erxx[0, j-y1]
            neweryy[:x1, j] = eryy[0, j-y1]
            newerzz[:x1, j] = erzz[0, j-y1]
            newer[:x1, j] = er[0, j-y1]
        for i in range(x1,Nx-x2):
            for j in range(y1,Ny-y2):
                newerxx[i, j] = erxx[i-x1, j-y1]
                neweryy[i, j] = eryy[i-x1, j-y1]
                newerzz[i, j] = erzz[i-x1, j-y1]
                newer[i, j] = er[i-x1, j-y1]

        newmetalnodesx=[(x1+i,y1+j) for (i,j) in metalnodesx]
        newmetalnodesy=[(x1+i,y1+j) for (i,j) in metalnodesy]
        newmetalnodesz=[(x1+i,y1+j) for (i,j) in metalnodesz]
        newcurrentloops=[[(a,x1+i,y1+j,b) for (a,i,j,b) in cloops] for cloops in currentloops]
        newvlines=[(a,x1+i,y1+j1,y1+j2) for (a,i,j1,j2) in vlines]

        self.dx = newdx
        self.dy = newdy
        self.dxcumsum = np.cumsum(self.dx)
        self.dycumsum = np.cumsum(self.dy)
        self.er = newer
        self.erxx = newerxx
        self.eryy = neweryy
        self.erzz = newerzz
        self.metalnodesx = newmetalnodesx
        self.metalnodesy = newmetalnodesy
        self.metalnodesz = newmetalnodesz
        self.vlines = newvlines
        self.currentloops = newcurrentloops
        # return (newdx, newdy, newer, newerxx, neweryy, newerzz, newmetalnodesx, newmetalnodesy, newmetalnodesz, newcurrentloops, newvlines)

    def sigma(self, i,dx,nlayer):
        """
        Returns conductivity value at a grid index. Returns conductivity associated with PML layer in PML regions. Returns 0 for indices outside PML region.

        Args:
            i(float): Index of point to get the conductivity at. Either full integer of half integer.
            dx(list): List of grid cell lengths including the PML region
            nlayer(list or tuple): 2-element container. First element is number of cells in PML region at the beginning of grid array. Second element if the number of cells in PML region at the end of grid array.

        Returns:
            Float: Conductivity value.
        """
        Rth=1.0e-7
        n=2.0
        n1=nlayer[0]
        n2=nlayer[1]
        if i>=n1 and i<=(len(dx)-n2):
            return 0
        elif i>=len(dx):
            return 0
        elif i<n1:
            delta1=sum(dx[:n1])
            dd=sum(dx[:int(i)])+(i-int(i))*dx[int(i)]
            return -(n+1)*epso*co/2.0/delta1*np.log(Rth)*np.power((delta1-dd)/delta1,n)
        elif i > (len(dx)-n2):
            delta2 = sum(dx[-n2:])
            dd = sum(dx[(len(dx)-n2):int(i)]) + (i - int(i)) * dx[int(i)]
            return -(n + 1) * epso * co / 2.0 / delta2 * np.log(Rth) * np.power(dd/ delta2, n)

    def sx(self,i):
        """
        Returns conductivity value at a dx grid index. Returns conductivity associated with PML layer in PML regions. Returns 0 for indices outside PML region.

        Args:
            i(float): Index of point to get the conductivity at. Either full integer of half integer.

        Returns:
            Float: Conductivity value.
        """
        return self.sigma(i,self.dx,self.pmllayers[:2])
    def sy(self,i):
        """
        Returns conductivity value at a dy grid index. Returns conductivity associated with PML layer in PML regions. Returns 0 for indices outside PML region.

        Args:
            i(float): Index of point to get the conductivity at. Either full integer of half integer.

        Returns:
            Float: Conductivity value.
        """
        return self.sigma(i,self.dy,self.pmllayers[-2:])

    def fdfd_solve(self):
        """Solves the FDFD problem. To reduce matrix dimensions, instead of classical equation of Ax=Lx, PQx=Lx formulation is used.
        """
        np.save(".\\temp\\"+"erxx.npy",self.erxx)
        np.save(".\\temp\\"+"eryy.npy",self.eryy)
        np.save(".\\temp\\"+"erzz.npy",self.erzz)
        np.save(".\\temp\\"+"metalnodesx.npy",self.metalnodesx)
        np.save(".\\temp\\"+"metalnodesy.npy",self.metalnodesy)
        np.save(".\\temp\\"+"metalnodesz.npy",self.metalnodesz)
        freq = self.freq
        dx = self.dx
        dy = self.dy
        er = self.er
        erxx = self.erxx
        eryy = self.eryy
        erzz = self.erzz
        metalnodesx = self.metalnodesx
        metalnodesy = self.metalnodesy
        metalnodesz = self.metalnodesz
        vlines = self.vlines
        cloops = self.currentloops
        pmllayers = self.pmllayers
        # sy.arr = dy
        # sx.arr = dx
        # sx.nlayer = pmllayers[:2]
        # sy.nlayer = pmllayers[-2:]
        import scipy.sparse
        # print("metalnodesx ",metalnodesx)
        # print("metalnodesy ",metalnodesy)
        # print("metalnodesz ",metalnodesz)
        # print("cloops ",cloops)
        # print("vline ",vline)

        ko = 2 * np.pi * freq / sabitler.speed_of_light

        # condition number düzeltme (bu yöntemin yanlış olduğuna karar verdim)
        #dd=np.min([np.min(dx),np.min(dy)])
        #coe = (1.0/ko/dd)/500
        #ko = ko*coe

        Nx = self.Nx
        Ny = self.Ny
        a = self.Nx * (self.Ny - 1)  # ExSayisi = HySayisi
        b = self.Ny * (self.Nx - 1)  # EySayisi = HxSayisi
        sparse= self.sparse
        print("sparse= "+str(sparse))
        if sparse:
            P = scipy.sparse.lil_matrix((a + b, a + b), dtype=np.complex128)
            Q = scipy.sparse.lil_matrix((a + b, a + b), dtype=np.complex128)
            # A = scipy.sparse.lil_matrix((2 * a + 2 * b, 2 * a + 2 * b), dtype=np.complex128)
            M1 = scipy.sparse.lil_matrix((a + b, a + b), dtype=np.complex128)
            # M1 = scipy.sparse.lil_matrix((2 * a + 2 * b, 2 * a + 2 * b), dtype=np.complex128)
        else:
            P = matlib.zeros((a + b, a + b), dtype=np.complex128)
            Q = matlib.zeros((a + b, a + b), dtype=np.complex128)
            # A = np.zeros((2 * a + 2 * b, 2 * a + 2 * b), dtype=np.complex128)
            M1 = matlib.zeros((a + b, a + b), dtype=np.complex128)
            # M1 = np.zeros((2 * a + 2 * b, 2 * a + 2 * b), dtype=np.complex128)

        dx = self.dx + [self.dx[-1]] + [self.dx[0]]  # bu sayede dx[-1] dx[0]'a esit olacak ve dy[Nx], dy[Nx-1]'e esit olacak.
        dy = self.dy + [self.dy[-1]] + [self.dy[0]]  # bu sayede dy[-1] dy[0]'a esit olacak ve dy[Ny], dy[Ny-1]'e esit olacak.
        ddx = [0.5 * (dx[i] + dx[i - 1]) for i in range(1, len(dx))]
        ddy = [0.5 * (dy[i] + dy[i - 1]) for i in range(1, len(dy))]
        dx = np.array(dx)
        dy = np.array(dy)

        for i in range(a + b):
            M1[i, i] = 1.0

        # def AF(x, y, z):
            # if y is not None:
                # A[x, y] = A[x, y] + z
        def PF(x, y, z):
            if y is not None:
                P[x, y] = P[x, y] + z
        def QF(x, y, z):
            if y is not None:
                Q[x, y] = Q[x, y] + z
        exn = self.exn
        eyn = self.eyn
        hxn = self.hxn
        hyn = self.hyn
        sx = self.sx
        sy = self.sy
        erxx = self.erxx
        eryy = self.eryy
        erzz = self.erzz
        for i in range(Nx):
            for j in range(1, Ny):
                row = exn(i, j)
                cc = c1 = c2 = c3 = c4 = 1.0
                if i>0:
                    c3 = (1.0/(1.0 - 1.0j * sx(i) / (ko * co * epso * 0.5 * (erxx[i, j] + erxx[i - 1, j]))))
                if i<Nx-1:
                    c1 = (1.0/ (1.0 - 1.0j * sx(i + 1.0) / (ko * co * epso * 0.5 * (erxx[i + 1, j] + erxx[i, j]))))
                    c2 = (1.0/ (1.0 - 1.0j * sy(j) / (ko * co * epso * 0.5 * (eryy[i + 1, j] + eryy[i + 1, j - 1]))))
                cc = (1.0/(1.0-1.0j*sx(i+0.5)/(ko*co*epso*erxx[i,j])))
                c4 = (1.0/(1.0-1.0j*sy(j)/(ko*co*epso*0.5*(eryy[i,j]+eryy[i,j-1]))))
                if row is not None:
                    if i > 0 and ((i, j) not in metalnodesz):  # Ez(i,j) katsayilari
                        PF(row, hxn(i, j - 1), -1.0 / erzz[i, j] / ko ** 2 / dx[i] / ddy[j - 1] *cc*c4)
                        PF(row, hxn(i, j), 1.0 / erzz[i, j] / ko ** 2 / dx[i] / ddy[j - 1] *cc*c4)
                        PF(row, hyn(i - 1, j), 1.0 / erzz[i, j] / ko ** 2 / dx[i] / ddx[i - 1] *cc*c3)
                        PF(row, hyn(i, j), -(1.0 / ddx[i - 1] / erzz[i, j]) / ko ** 2 / dx[i] *cc*c3)

                    if i < Nx - 1 and ((i + 1, j) not in metalnodesz):  # Ez(i+1,j) katsayilari
                        PF(row, hxn(i + 1, j), -1.0 / erzz[i + 1, j] / ko ** 2 / dx[i] / ddy[j - 1] *cc*c2)
                        PF(row, hxn(i + 1, j - 1), 1.0 / erzz[i + 1, j] / ko ** 2 / dx[i] / ddy[j - 1] *cc*c2)
                        PF(row, hyn(i + 1, j), 1.0 / erzz[i + 1, j] / ko ** 2 / dx[i] / ddx[i] *cc*c1)
                        PF(row, hyn(i, j), -(1.0 / ddx[i] / erzz[i + 1, j]) / ko ** 2 / dx[i] *cc*c1)
                    PF(row, hyn(i, j), 1.0)

        for i in range(1, Nx):
            for j in range(Ny):
                row = eyn(i, j)
                cc = c1 = c2 = c3 = c4 = 1.0
                if j > 0:
                    c4 = (1.0/ (1.0 - 1.0j * sy(j) / (ko * co * epso * 0.5 * (eryy[i, j] + eryy[i, j - 1]))))
                if j < Ny-1:
                    c1 = (1.0/ (1.0 - 1.0j * sx(i) / (ko * co * epso * 0.5 * (erxx[i, j+1] + erxx[i-1, j+1]))))
                    c2 = (1.0/ (1.0 - 1.0j * sy(j+1) / (ko * co * epso * 0.5 * (eryy[i, j+1] + eryy[i, j]))))
                cc = (1.0/ (1.0 - 1.0j * sy(j + 0.5) / (ko * co * epso * eryy[i, j])))
                c3 = (1.0/ (1.0 - 1.0j * sx(i) / (ko * co * epso * 0.5 * (erxx[i, j] + erxx[i - 1, j]))))
                if row is not None:
                    if j > 0 and ((i, j) not in metalnodesz):  # Ez(i,j) katsayilari
                        PF(row, hyn(i - 1, j), 1.0 / erzz[i, j] / ko ** 2 / ddx[i - 1] / dy[j] *cc*c3)
                        PF(row, hyn(i, j), -1.0 / erzz[i, j] / ko ** 2 / ddx[i - 1] / dy[j] *cc*c3)
                        PF(row, hxn(i, j - 1), -1.0 / erzz[i, j] / ko ** 2 / ddy[j - 1] / dy[j] *cc*c4)
                        PF(row, hxn(i, j), (1.0 / ddy[j - 1] / erzz[i, j]) / ko ** 2 / dy[j] *cc*c4)
                    if j < Ny - 1 and ((i, j + 1) not in metalnodesz):  # Ez(i,j+1) katsayilari
                        PF(row, hyn(i - 1, j + 1), -1.0 / erzz[i, j + 1] / ko ** 2 / ddx[i - 1] / dy[j] *cc*c1)
                        PF(row, hyn(i, j + 1), 1.0 / erzz[i, j + 1] / ko ** 2 / ddx[i - 1] / dy[j] *cc*c1)
                        PF(row, hxn(i, j + 1), -1.0 / erzz[i, j + 1] / ko ** 2 / ddy[j] / dy[j] *cc*c2)
                        PF(row, hxn(i, j), (1.0 / ddy[j] / erzz[i, j + 1]) / ko ** 2 / dy[j] *cc*c2)
                    PF(row, hxn(i, j), -1.0)

        for i in range(Nx):
            for j in range(Ny):
                row = hyn(i, j)
                cc = c1 = c2 = c3 = c4 = 1.0
                if j > 0:
                    c4 = (1.0/ (1.0 - 1.0j * sy(j - 0.5) / (ko * co * epso * er[i, j-1])))
                    cc = (1.0/ (1.0 - 1.0j * sy(j) / (ko * co * epso * 0.5 * (eryy[i, j]+eryy[i, j-1]))))
                    c3 = (1.0/ (1.0 - 1.0j * sx(i + 0.5) / (ko * co * epso * er[i, j-1])))
                c1 = (1.0/ (1.0 - 1.0j * sx(i + 0.5) / (ko * co * epso * er[i, j])))
                c2 = (1.0/ (1.0 - 1.0j * sy(j + 0.5) / (ko * co * epso * er[i, j])))
                if row is not None:
                    QF(row, eyn(i, j - 1), -1.0 / ko ** 2 / dx[i] / (0.5 * (dy[j] + dy[j - 1])) *cc*c3)
                    QF(row, eyn(i, j), 1.0 / ko ** 2 / dx[i] / (0.5 * (dy[j] + dy[j - 1])) *cc*c1)
                    QF(row, eyn(i + 1, j), -1.0 / ko ** 2 / dx[i] / (0.5 * (dy[j] + dy[j - 1])) *cc*c1)
                    QF(row, eyn(i + 1, j - 1), 1.0 / ko ** 2 / dx[i] / (0.5 * (dy[j] + dy[j - 1])) *cc*c3)
                    QF(row, exn(i, j - 1), 1.0 / ko ** 2 / dy[j - 1] / (0.5 * (dy[j] + dy[j - 1])) *cc*c4)
                    QF(row, exn(i, j + 1), 1.0 / ko ** 2 / dy[j] / (0.5 * (dy[j] + dy[j - 1])) *cc*c2)
                    QF(row, exn(i, j),
                       erxx[i, j] - (1.0 / dy[j] *cc*c2 + 1.0 / (dy[j - 1]) *cc*c4) / ko ** 2 / (0.5 * (dy[j] + dy[j - 1])))

        for i in range(Nx):
            for j in range(Ny):
                row = hxn(i, j)
                cc = c1 = c2 = c3 = c4 = 1.0
                if i > 0:
                    c4 = (1.0/ (1.0 - 1.0j * sy(j + 0.5) / (ko * co * epso * er[i-1, j])))
                    cc = (1.0/ (1.0 - 1.0j * sx(i) / (ko * co * epso * 0.5 * (erxx[i, j]+erxx[i-1, j]))))
                    c3 = (1.0/ (1.0 - 1.0j * sx(i - 0.5) / (ko * co * epso * er[i-1, j])))
                c1 = (1.0/ (1.0 - 1.0j * sx(i + 0.5) / (ko * co * epso * er[i, j])))
                c2 = (1.0/ (1.0 - 1.0j * sy(j + 0.5) / (ko * co * epso * er[i, j])))
                if row is not None:
                    QF(row, exn(i - 1, j), 1.0 / ko ** 2 / (0.5 * (dx[i] + dx[i - 1])) / dy[j] *cc*c4)
                    QF(row, exn(i, j), -1.0 / ko ** 2 / (0.5 * (dx[i] + dx[i - 1])) / dy[j] *cc*c2)
                    QF(row, exn(i - 1, j + 1), -1.0 / ko ** 2 / (0.5 * (dx[i] + dx[i - 1])) / dy[j] *cc*c4)
                    QF(row, exn(i, j + 1), 1.0 / ko ** 2 / (0.5 * (dx[i] + dx[i - 1])) / dy[j] *cc*c2)
                    QF(row, eyn(i - 1, j), -1.0 / ko ** 2 / (0.5 * (dx[i] + dx[i - 1])) / dx[i - 1] *cc*c3)
                    QF(row, eyn(i + 1, j), -1.0 / ko ** 2 / (0.5 * (dx[i] + dx[i - 1])) / dx[i] *cc*c1)
                    QF(row, eyn(i, j),
                       -eryy[i, j] + (1.0 / dx[i] *cc*c1 + 1.0 / dx[i - 1] *cc*c3) / ko ** 2 / (0.5 * (dx[i] + dx[i - 1])))

        coef=1e30
        for i, j in metalnodesx:
            r = exn(i, j)
            P[r, r] = P[r, r] * coef
            Q[r, r] = Q[r, r] * coef
            M1[r, r] = M1[r, r] * coef
        for i, j in metalnodesy:
            r = eyn(i, j)
            P[r, r] = P[r, r] * coef
            Q[r, r] = Q[r, r] * coef
            M1[r, r] = M1[r, r] * coef
        # for i in range(2 * a + 2 * b):
        #     for j in range(2 * a + 2 * b):
        #         A[i, j] = A[i, j]/100.0
        #         M1[i, j] = M1[i, j]/100.0
        import scipy.sparse.linalg
        from numpy import linalg as LA
        import scipy.linalg
        # A = A + coe*np.eye(2 * a + 2 * b)
        if sparse:
            la, v = scipy.sparse.linalg.eigs(P*Q, k=100, M=M1, sigma=np.mean(er), tol=0.001, which="LM")
        else:
            cond_number=np.linalg.cond(P*Q)
            print("condition number= ",cond_number)
            la, v = scipy.linalg.eig(P*Q, M1)
        # Field Vector: [Ex Ey]
        # eigenvalue, eigenvector pairs
        eigs = []
        for i in range(len(la)):
            print("la[i] ",la[i])
            if abs(la[i].imag / la[i].real) < 0.02 and la[i].real > 0.1  and la[i].real < 2*np.max(er):
            # if abs(la[i].imag) < 1e-2 and la[i].real > 0.1 and abs(la[i].real)**2 < np.max(er)*1.2:
                print("in_la[i] ",la[i])
                eigs.append((la[i], v[:, i]))

        output = []
        eyn= self.eyn
        hyn= self.hyn
        hxn= self.hxn
        for la, Ev in eigs:
            # Length of eigs can be higher than 1, so choose appropriate mode.
            currents = []
            voltages = []
            ko = 2 * np.pi * freq / co
            eeff = la
            lambda_g = 2 * np.pi / np.sqrt(eeff) / ko
            Hv = Q*np.matrix(Ev).transpose()/np.sqrt(la+0j)
            Hv,=np.array(Hv.T)
            mydict = locals().copy() # we use local scope, because eyn is defined in local scope
                                     # and I cannot use "self." in eval.
            for cloop in cloops:
                print("cloop= ",cloop)
                temp=[(eval("{}({},{})".format(func,i,j),mydict),dl) for func, i, j, dl in cloop]
                cf = [Hv[i]*j for i,j in temp if i is not None]
                currents.append( np.sum(cf))
            for cc in range(len(vlines)):
                func, i0, j1, j2 = vlines[cc]
                temp=[(eval("{}({},{})".format(func,i0,j),mydict),dy[j]) for j in range(j1,j2)]
                voltages.append( np.sum([Ev[i]*j for i,j in temp if i is not None]))

            Z1 = np.abs(np.sqrt((sabitler.mu_0/ sabitler.epsilon_0)) * np.array(voltages) / np.array(currents))
            # P = powerflow(dx[1:-1], dy[1:-1], (Ev,Hv))
            P = self.powerflow((Ev,Hv))
            print("Z1= "+str(Z1))
            print("P= "+str(P))
            Z2 = np.sqrt((sabitler.mu_0/ sabitler.epsilon_0)) *np.abs(np.array(voltages))**2/P
            print("eeff ",eeff)
            print("Z2= "+str(Z2))
            print(Z2)
            Z3 = np.max(np.abs(Ev))/np.max(np.abs(Hv))
            print("Z3= "+str(Z3))
            print(Z3)
            print("voltage")
            print(np.abs(np.array(voltages)))
            print("maxEv")
            print(np.abs(np.max(Ev)))

            # import xlsxwriter
            #workbook = xlsxwriter.Workbook('Ey.xlsx')
            #worksheet = workbook.add_worksheet()
            #for i in range(len(dx)):
            #    worksheet.write(0,i, str(i))
            #    for j in range(len(dy)):
            #        if eyn(i,j):
            #            worksheet.write(j+1,i, str(np.abs(Ev[eyn(i,j)])).replace(".",","))
            #workbook.close()
            # workbook = xlsxwriter.Workbook('Hx.xlsx')
            # worksheet = workbook.add_worksheet()
            # for i in range(len(dx)):
                # worksheet.write(0,i, str(i))
                # for j in range(len(dy)):
                    # if hxn(i,j):
                        # worksheet.write(j+1,i, str(np.abs(Hv[hxn(i,j)])).replace(".",","))
            # workbook.close()

            output.append((Z2, eeff, lambda_g, la, (Ev,Hv) ))
        # self.finishedsim.emit(output)
        return output

    @classmethod
    def microstrip(cls, boxwidth, boxheight, dielheight, metalwidth, metalthickness, epsilon, freq, xres, yres):
        pass

    @classmethod
    def stripline(cls, boxwidth, boxheight, metalwidth, metalthickness, epsilon, freq, xres, yres):
        pass

    @classmethod
    def suspended(cls, boxwidth, boxheight, dielthickness, metalheight, metalwidth, metalthickness, epsilon, freq, xres, yres):
        pass

    @classmethod
    def multilayer(cls, boxwidth, boxheight, dielthickness, epsilon, metals, freq, pmllayers, xres, yres):
        """
        Multilayer substrate.

        Args:
            dielthickness (list): tabandan başlayarak yukarı dogru sirasiyla dielektriklerin kalinliklarinin listesi
            epsilon (list): tabandan başlayarak yukarı dogru sirasiyla dielektriklerin katsayilarinin listesi
            metals (list):şu elemanlardan oluşan dizi [metalheight, metalwidth, metalthickness, metaloffset]
            metalheight (float): metalin alt yüzeyinin yüksekliği
        """
        obj = cls()
        obj.freq = freq
        obj.boxwidth = boxwidth
        obj.boxheight = boxheight
        # resolution = boxwidth/50.0
        # print("yres= "+str(yres))
        # if yres == None:
            # yres = resolution
        # if xres == None:
            # xres = resolution
        # parts = []
        temp = 0
        for diel, eps in zip(dielthickness, epsilon):
            obj.parts.append(("dielectric", "rect", (boxwidth/ 2.0), temp + (diel/ 2.0), boxwidth, diel, eps))
            temp = temp + diel
        for metal in metals:
            metalheight, metalwidth, metalthickness, metaloffset = tuple(metal)
            obj.parts.append(("metal", "rect", (boxwidth/2.0)+metaloffset, metalheight + (metalthickness/ 2.0), metalwidth, metalthickness))
        obj.fdfd_mesh(xres, yres)
        obj.addpml(pmllayers)
        return obj
        # cc = getindice((boxwidth/ 2.0), dx, pmllayers[0])
        # print("cc= ", cc, getindice(metalheight, dy, pmllayers[2]))
        # vline = [(eyn(cc, i), dy[i]) for i in range(getindice(metalheight, dy, pmllayers[2]))]
        # return (dx, dy, er, erxx, eryy, erzz, metalnodesx, metalnodesy, metalnodesz, cloops, vlines, parts)

    @classmethod
    def waveguide(cls, boxwidth, boxheight, dielthickness, epsilon, freq, pmllayers, xres, yres):
        """Generates rectangular waveguide problem. Multilayer function is not used because of the definition of voltage line wihtout any metal.

        Args:
            dielthickness (list): tabandan başlayarak yukarı dogru sirasiyla dielektriklerin kalinliklarinin listesi
            epsilon (list): tabandan başlayarak yukarı dogru sirasiyla dielektriklerin katsayilarinin listesi
            metalwidth (float): object
        """
        obj = cls()
        obj.freq = freq
        obj.boxwidth = boxwidth
        obj.boxheight = boxheight
        # resolution = np.min([(metalwidth/ 10.0), co / freq / np.sqrt(np.max(epsilon)) / 20.0, (np.min([boxheight,boxwidth])/40.0)])
        # resolution = (metalthickness/ 2.0)
        # if yres == None:
            # yres = resolution
        # if xres == None:
            # xres = resolution
        # parts = []
        temp = 0
        for diel, eps in zip(dielthickness, epsilon):
            #parts.append(("dielectric", "rect", boxwidth / 2.0, temp + diel / 2.0, boxwidth, diel, eps))
            obj.parts.append(("dielectric", "rect", boxwidth / 2.0, temp + diel / 2.0, boxwidth, diel, eps))
        obj.fdfd_mesh(xres,yres)
        obj.addpml(pmllayers)
        # cc = getindice((boxwidth/ 2.0), obj.dx, pmllayers[0])
        cc = getindiceordered((boxwidth/ 2.0), obj.dxcumsum, pmllayers[0])
        #vline = [(eyn(cc, i), dy[i]) for i in range(getindice(boxheight, dy, pmllayers[2]))]
        # obj.vlines = [("eyn",cc, 0, getindice(boxheight, obj.dy, pmllayers[2])+1)]
        obj.vlines = [("eyn",cc, 0, getindiceordered(boxheight, obj.dycumsum, pmllayers[2])+1)]
        return obj
        # print("cloops1= ")
        # print(cloops)
        # return (dx, dy, er, erxx, eryy, erzz, metalnodesx, metalnodesy, metalnodesz, cloops, vlines, parts)

    def plot_parts(self, ekran, solid=False, shift=(0,0)):
        """

        Args:
            ekran: Matplotlib Axes object
            solid(bool): if True, use fill=true, else fill=false
        """
        parts = self.parts
        # ekran.add_patch(Rectangle((0.0, 0.0), p[4], p[5], fc='r'))
        for p in parts:
            if p[1] == "rect":
                if p[0] == "dielectric":
                    xo = p[2] - (p[4]/ 2.0)
                    yo = p[3] - (p[5]/ 2.0)
                    ekran.axes.add_patch(mpatches.Rectangle((xo+shift[0], yo+shift[1]), p[4], p[5], fill=solid, fc='green'))
                elif p[0] == "metal":
                    xo = p[2] - (p[4]/ 2.0)
                    yo = p[3] - (p[5]/ 2.0)
                    ekran.axes.add_patch(mpatches.Rectangle((xo+shift[0], yo+shift[1]), p[4], p[5], fill=solid, fc='red'))

    def plot_pml(self, ekran, solid=False, shift=(0,0)):
        """

        Args:
            ekran: Matplotlib Axes object
            solid(bool): if True, use fill=true, else fill=false
        """
        pmllayers = self.pmllayers
        dx = self.dx
        dy = self.dy
        # ekran.add_patch(Rectangle((0.0, 0.0), p[4], p[5], fc='r'))
        x1,x2,y1,y2=tuple(pmllayers)
        lx=np.sum(dx[:])
        ly=np.sum(dy[:])
        if x1>0:
            ekran.axes.add_patch(mpatches.Rectangle((0, 0), np.sum(dx[:x1]),ly, fill=solid, fc='blue'))
        if x2>0:
            ekran.axes.add_patch(mpatches.Rectangle((lx-np.sum(dx[-x2:]), 0), np.sum(dx[-x2:]),ly, fill=solid, fc='blue'))
        if y1>0:
            ekran.axes.add_patch(mpatches.Rectangle((0, 0), lx, np.sum(dy[:y1]), fill=solid, fc='blue'))
        if y2>0:
            ekran.axes.add_patch(mpatches.Rectangle((0, ly-np.sum(dy[-y2:]), 0), lx, np.sum(dy[-y2:]), fill=solid, fc='blue'))

    def plot_vline(self, ekran):
        dx = self.dx
        dy = self.dy
        vlines = self.vlines
        for vline in vlines:
            func,i,j1,j2 = vline
            print(f"vline {i} {j1} {j2}")
            if func=="eyn":
                xo = np.sum(dx[:i])
                y1=np.sum(dy[:j1])
                y2=np.sum(dy[:j2])
                #ekran.axes.plot([xo, xo], [y1, y2], "red")
                ekran.axes.arrow(xo, y1, 0, y2-y1, width=dx[i]/2,color="red",length_includes_head=True)

    def plot_cloops(self, ekran):
        dx = self.dx
        dy = self.dy
        cloops = self.currentloops
        for cloop in cloops:
            for loop in cloop:
                func,i,j,dd = loop
                # print(f"loop {i} {j} {dd}")
                px=[]
                py=[]
                if func=="hxn":
                    yo = np.sum(dy[:j+1])-dy[j]/2
                    x1 = np.sum(dx[:i])-dx[i-1]/2
                    x2 = np.sum(dx[:i])+dx[i]/2
                    px.append(x1)
                    px.append(x2)
                    py.append(yo)
                    py.append(yo)
                    if dd<0:
                        px.reverse()
                elif func=="hyn":
                    xo = np.sum(dx[:i+1])-dx[i]/2
                    y1 = np.sum(dy[:j])-dy[j-1]/2
                    y2 = np.sum(dy[:j])+dy[j]/2
                    px.append(xo)
                    px.append(xo)
                    py.append(y1)
                    py.append(y2)
                    if dd<0:
                        py.reverse()
                #ekran.axes.plot(px, py, "blue")
                ekran.axes.arrow(px[0], py[0], px[1]-px[0], py[1]-py[0], width=np.abs(dd)/10,color="blue",length_includes_head=True)

    def plot_grids(self, ekran):
        dx = self.dx
        dy = self.dy
        boxwidth = sum(dx)
        boxheight = sum(dy)
        xp = dx.copy()
        yp = dy.copy()
        for i in range(1, len(dx)):
            xp[i] = xp[i] + xp[i - 1]
        for i in range(1, len(dy)):
            yp[i] = yp[i] + yp[i - 1]
        xp = [0.0] + xp
        yp = [0.0] + yp
        for xo in xp:
            ekran.axes.plot([xo, xo], [0.0, boxheight], "black")
        for yo in yp:
            ekran.axes.plot([0.0, boxwidth], [yo, yo], "black")

    def powerflow(self, v):
        """

        Args:
            comp: "Ex","Ey"...
            surface: if 1, use contourf, else streamplot
            v: (Ev, Hv)
        """
        dx = self.dx
        dy = self.dy
        eyn= self.eyn
        exn= self.exn
        hyn= self.hyn
        hxn= self.hxn
        boxwidth = sum(dx)
        boxheight = sum(dy)
        nx=100
        ny=100
        xi = np.linspace(0, boxwidth, nx)
        yi = np.linspace(0, boxheight, ny)
        xp = dx.copy()
        yp = dy.copy()
        for i in range(1, len(dx)):
            xp[i] = xp[i] + xp[i - 1]
        for i in range(1, len(dy)):
            yp[i] = yp[i] + yp[i - 1]
        xtemp = []
        ytemp = []
        Ey = []
        Hx=[]
        for j in range(len(dy)):
            xtemp += [0.0, boxwidth]
            ytemp += [yp[j] - (dy[j]/ 2.0), yp[j] - (dy[j]/ 2.0)]
            Ey += [0.0, 0.0]
            Hx += [0.0, 0.0]
            for i in range(1, len(dx)):
                xtemp.append(xp[i - 1])
                ytemp.append(yp[j] - (dy[j]/ 2.0))
                Ey.append(v[0][eyn(i, j)])
                Hx.append(v[1][hxn(i, j)])
        Eyy = griddata((xtemp, ytemp), Ey, (xi[None, :], yi[:, None]), method='nearest')
        Hxx = griddata((xtemp, ytemp), Hx, (xi[None, :], yi[:, None]), method='nearest')
        xtemp = []
        ytemp = []
        Ex = []
        Hy = []
        TEx=[]
        for i in range(len(dx)):
            xtemp += [xp[i] - (dx[i]/ 2.0), xp[i] - (dx[i]/ 2.0)]
            ytemp += [0.0, boxheight]
            Ex += [0.0, 0.0]
            Hy += [0.0, 0.0]
            for j in range(1, len(dy)):
                xtemp.append(xp[i] - (dx[i]/ 2.0))
                ytemp.append(yp[j - 1])
                Ex.append(v[0][exn(i, j)])
                Hy.append(v[1][hyn(i, j)])
        Exx = griddata((xtemp, ytemp), Ex, (xi[None, :], yi[:, None]), method='nearest')
        Hyy = griddata((xtemp, ytemp), Hy, (xi[None, :], yi[:, None]), method='nearest')
        #magE = np.sqrt(np.abs(Exx * Exx + Eyy * Eyy))
        #magH = np.sqrt(np.abs(Hxx * Hxx + Hyy * Hyy))
        print(np.shape(Exx))
        print(np.shape(Hyy))
        print(Exx)
        print(Hyy)
        a1 = np.trapz(np.trapz(np.real(Exx*np.conjugate(Hyy)))*boxwidth/(nx-1))*boxheight/(ny-1)
        a2 = np.trapz(np.trapz(np.real(Eyy*np.conjugate(Hxx)))*boxwidth/(nx-1))*boxheight/(ny-1)
        return np.abs(a2-a1)

    def plot_field(self, ekran, v, surface=True, comp="Ex"):
        """

        Args:
            surface(bool): if True, use contourf, else streamplot
            comp(str): "Ex","Ey"...
        """
        axes = ekran.axes
        figure = ekran.figure
        eyn= self.eyn
        exn= self.exn
        hyn= self.hyn
        hxn= self.hxn

        # figure = plt.figure()
        # axes=figure.add_subplot(111)
        dx = self.dx
        dy = self.dy
        parts = self.parts
        pmllayers = self.pmllayers
        boxwidth = sum(dx)
        boxheight = sum(dy)
        xi = np.linspace(0, boxwidth, 100)
        yi = np.linspace(0, boxheight, 100)
        xp = dx.copy()
        yp = dy.copy()
        for i in range(1, len(dx)):
            xp[i] = xp[i] + xp[i - 1]
        for i in range(1, len(dy)):
            yp[i] = yp[i] + yp[i - 1]
        xtemp = []
        ytemp = []
        Ey = []
        Hx=[]
        for j in range(len(dy)):
            xtemp += [0.0, boxwidth]
            ytemp += [yp[j] - (dy[j]/ 2.0), yp[j] - (dy[j]/ 2.0)]
            Ey += [0.0, 0.0]
            Hx += [0.0, 0.0]
            for i in range(1, len(dx)):
                xtemp.append(xp[i - 1])
                ytemp.append(yp[j] - (dy[j]/ 2.0))
                Ey.append(v[0][eyn(i, j)])
                Hx.append(v[1][hxn(i, j)])
                # Constant extrapolation of fields to points on boundary
                if j==len(dy)-1:
                    xtemp.append(xp[i - 1])
                    ytemp.append(boxheight)
                    Ey.append(v[0][eyn(i, j)])
                    Hx.append(v[1][hxn(i, j)])
                if j==0:
                    xtemp.append(xp[i - 1])
                    ytemp.append(0)
                    Ey.append(v[0][eyn(i, j)])
                    Hx.append(v[1][hxn(i, j)])
        xtemp.extend([0,0,boxwidth,boxwidth])
        ytemp.extend([0,boxheight,boxheight,0])
        Ey.extend([0,0,0,0])
        Hx.extend([0,0,0,0])
        Eyy = griddata((xtemp, ytemp), Ey, (xi[None, :], yi[:, None]), method='cubic')
        Hxx = griddata((xtemp, ytemp), Hx, (xi[None, :], yi[:, None]), method='cubic')
        xtemp = []
        ytemp = []
        Ex = []
        Hy = []
        for i in range(len(dx)):
            xtemp += [xp[i] - (dx[i]/ 2.0), xp[i] - (dx[i]/ 2.0)]
            ytemp += [0.0, boxheight]
            Ex += [0.0, 0.0]
            Hy += [0.0, 0.0]
            for j in range(1, len(dy)):
                xtemp.append(xp[i] - (dx[i]/ 2.0))
                ytemp.append(yp[j - 1])
                Ex.append(v[0][exn(i, j)])
                Hy.append(v[1][hyn(i, j)])
                # Constant extrapolation of fields to points on boundary
                if i==0:
                    xtemp.append(0)
                    ytemp.append(yp[j - 1])
                    Ex.append(v[0][exn(i, j)])
                    Hy.append(v[1][hyn(i, j)])
                if i==len(dx)-1:
                    xtemp.append(boxwidth)
                    ytemp.append(yp[j - 1])
                    Ex.append(v[0][exn(i, j)])
                    Hy.append(v[1][hyn(i, j)])
        xtemp.extend([0,0,boxwidth,boxwidth])
        ytemp.extend([0,boxheight,boxheight,0])
        Ex.extend([0,0,0,0])
        Hy.extend([0,0,0,0])

        Exx = griddata((xtemp, ytemp), Ex, (xi[None, :], yi[:, None]), method='cubic')
        Hyy = griddata((xtemp, ytemp), Hy, (xi[None, :], yi[:, None]), method='cubic')
        magE = np.sqrt(np.abs(Exx * Exx + Eyy * Eyy))
        magH = np.sqrt(np.abs(Hxx * Hxx + Hyy * Hyy))

        if surface:
            Quantity=[]
            if comp == "Ex":
                Quantity = np.abs(Exx)
                nz.autoscale(magE)
            elif comp == "Ey":
                Quantity = np.abs(Eyy)
                nz.autoscale(magE)
            elif comp == "E_t":
                Quantity = np.abs(magE)
                nz.autoscale(magE)
            elif comp == "Hx":
                Quantity = np.abs(Hxx)
                nz.autoscale(magH)
            elif comp == "Hy":
                Quantity = np.abs(Hyy)
                nz.autoscale(magH)
            elif comp == "H_t":
                Quantity = np.abs(magH)
                nz.autoscale(magH)
            X, Y = np.meshgrid(xi, yi)
            temp = axes.contourf(X, Y, Quantity, color='black', density=10.0)
            if self.plot_field_cb:
                self.plot_field_cb.remove()
            self.plot_field_cb = figure.colorbar(temp)
        else:
            if comp=="E_t":
                nz.autoscale(magE)
                temp = axes.streamplot(xi, yi, np.real(Exx), np.real(Eyy), color=magE, density=2.0, cmap=cm.gist_rainbow,
                                  linewidth=1, arrowstyle='->')
            elif comp == "H_t":
                nz.autoscale(magH)
                temp = axes.streamplot(xi, yi, np.real(Hxx), np.real(Hyy), color=magH, density=2.0, cmap=cm.gist_rainbow,
                                          linewidth=1, arrowstyle='->')    # ekran.axes.colorbar()
            # if self.plot_field_cb:
                # self.plot_field_cb.ax.clear()
                # self.plot_field_cb = ekran.figure.colorbar(temp.lines, cax = self.plot_field_cb.ax)
            # else:
                # self.plot_field_cb = ekran.figure.colorbar(temp.lines)


if __name__ == "__main__":
    #import time
    dd = np.ones(200)
    ddcs = np.cumsum(dd)
    cc=195.5
    print(ddcs)
    import time
    t1 = time.time()
    # for i in range(10000):
    print(getindice(cc,dd))
    t2 = time.time()
    # for i in range(10000):
    print(getindiceordered(cc,ddcs))
    t3 = time.time()
    print(t2-t1)
    print(t3-t2)
    sys.exit(0)
    #t1 = time.time()
    #for i in range(5000):
    #    getindice(4.9,dd)
    #t2 = time.time()
    #for i in range(5000):
    #    getindiceordered(4.9,ddcs)
    #t3 = time.time()
    #print(t2-t1)
    #print(t3-t2)
    #sys.exit(0)
    boxwidth = 3.1e-3  # WG genişliği
    boxheight = 1.55e-3 # WG yüksekliği
    dielheight = 1.55e-3
    metalwidth = 0.01
    metalthickness = 0.002
    freq = 77e9
    epsilon = 1.0
    pml_layers=[0,0,0,0]
    # ms = microstrip(boxwidth, boxheight, dielheight, metalwidth, metalthickness, epsilon, freq)
    #ms = multilayer(boxwidth, boxheight, [dielheight], [epsilon], [[dielheight, metalwidth, metalthickness,0.0]], freq, pml_layers,0.001,0.002)
    ms = waveguide(boxwidth, boxheight, [dielheight], dielheight, metalwidth, metalthickness, [epsilon], freq, pml_layers,0.000025,0.000025)
    #fig = plt.figure()
    out = fdfd_solve(*(list(ms[:-1]) + [freq]+[pml_layers]))
    #plot_field(fig.gca(), ms[0], ms[1], out[0][-1], ms[-1],pml_layers, surface=True,comp="Ey")
    # plot_field(fig.gca(), ms[0], ms[1], out[0][-1], ms[-1],pml_layers, surface=True,comp="H_t")
    #plot_grids(fig.gca(), ms[0], ms[1])
    #plot_parts(fig.gca(), ms[-1])
    print(out[0][0],out[0][1])
    ko=2*np.pi*freq/3e8
    kc=np.pi/boxwidth
    beta = np.sqrt(-kc**2+ko**2)
    print(beta/ko)
    print(2*np.pi/beta)
    #print(2*np.pi/(np.sqrt(np.real(out[0][0]))*ko))
    print(out[0][2])
    # print "eigs "
    # for i in out:
    #     print "i= ", i[0], i[1]
    # print ms[-2]

    # exn.Nx = eyn.Nx = hxn.Nx = hyn.Nx = Nx = 3
    # exn.Ny = eyn.Ny = hxn.Ny = hyn.Ny = Ny = 2
    # ind={}
    # for i in range(Nx+1):
        # for j in range(Ny+1):
            # print "exnnn ",i,j,exn.Nx,exn.Ny
            # if exn(i,j) is not None::
                # print "exn ",i,j,exn(i,j)
    # for i in range(Nx+1):
        # for j in range(Ny+1):
            # if eyn(i,j) is not None::
                # print "eyn ",i,j,eyn(i,j)
    # for i in range(Nx+1):
        # for j in range(Ny+1):
            # if hxn(i,j) is not None::
                # print "hxn ",i,j,hxn(i,j)
    # for i in range(Nx+1):
        # for j in range(Ny+1):
            # if hyn(i,j) is not None::
                # print "hyn ",i,j,hyn(i,j)

    # plt.show()



import sys
sys.path.append(r"c:\\users\\erdoel\\documents\\works\\python_works\\microwave_toolbox")
from mwtoolbox.rfnetwork import *
import network
import itertools
import numpy as np
frequencies=np.linspace(74e9,83e9,8,endpoint=True)

sp = spfile("WR12_SIW_MS_Transition_Design1_29.s2p")
# sp = spfile("WR12_SIW_MS_Transition_Design1_Al_dB_unnormalized.s2p")
# sp = spfile("WR12_SIW_MS_Transition_Design1_Al_RI_unnormalized.s2p")
print(sp.params)
# print(sp.gammas)
sp.set_frequency_points(frequencies)

import matplotlib.pyplot as plt

# plt.plot(frequencies,sp.S(1,1,"dB"))
# plt.plot(frequencies,sp.S(2,2,"dB"))
# plt.grid()

# plt.figure()
# plt.plot(frequencies,sp.S(2,1,"dB"))
# plt.grid()
# plt.show()

# plt.plot(frequencies,np.real(sp.gammas[0]),"+")
# frequencies=np.linspace(74e9,83e9,9,endpoint=True)
# sp.set_frequency_points(frequencies)
# plt.plot(frequencies,np.real(sp.gammas[0]),"*")
# plt.grid()
# plt.show()



#-*-coding:utf-8-*-
from .myconstants import *
from .components import z_wg_te10
from scipy import sin,sinh,arcsin,arcsinh,cos,cosh,arccos,arccosh,tan,arctan,tanh,arctanh,log,log10,power
from scipy.special import *
from scipy.optimize import brentq
from scipy.integrate import quad
from string import *
import re
import numpy as np
from numpy.lib.scimath import sqrt as csqrt
from .network import *
co=speed_of_light_in_freespace.simplified.magnitude
eta0=free_space_wave_impedance.simplified.magnitude
mu0=free_space_permeability.simplified.magnitude
eps0=free_space_permittivity.simplified.magnitude

def zwg_te10(er,a,b,freq):
    return z_wg_te10(er,a,b,freq,1)

def ewg_abcd(a,b,er,length,frek):
    #Referans:"The Design of Evanescent Mode Waveguide Bandpass Filters for a Prescribed Insertion Loss Characteristic.pdf"
    # Model= Xp1,Xs1,Xp1 ya da Xs2,Xp2,Xs2 (p: shunt, s: series)
    #Zo=jXo
    lcutoff=2*a
    wavelength=co/csqrt(er)/frek
    Xo=abs(zwg_te10(er,a,b,frek))
    gamma=2.*pi/wavelength*csqrt(((wavelength/lcutoff))**2.0-1.)
    Xs1=Xo*sinh(gamma*length)
    Xs2=Xo*tanh(gamma*length/2.0)
    Xp1=(Xo/tanh(gamma*length/2.0))
    Xp2=(Xo/sinh(gamma*length))
    networks=[]
    #networks.append(shunt_z(-1.0j*Xo*tanh(gamma*length)))
    networks.append(shunt_z(1.0j*Xp1))
    networks.append(series_z(1.0j*Xs1))
    networks.append(shunt_z(1.0j*Xp1))
    #networks.append(shunt_z(-1.0j*Xo*tanh(gamma*length)))
    return cascade_networks(networks)

def minimum_butterworth_filter_degree(L,fstop):
    #KAYNAK:Microstrip Filters for RF Microwave Applications s.42, fstop: normalize frekans, L: fstoptaki bastirma (dB)
    return int((log10(pow(10.0,0.1*L)-1.0)/(2.0*log10(fstop))))+1

def butterworth_filter_prototype(n):
    #KAYNAK:Microstrip Filters for RF Microwave Applications s.41, fstop'taki bastirmayi da soyler, fstop: normalize frekans
    g=list(range(n+2))
    g[0]=g[n+1]=1.0
    g[1:n+1]=2.*sin((2*np.array(list(range(1, n+1)))-1)*pi/2./n)
    return g, -10.0*log10(1.0+pow(fstop,2*n))

def minimum_chebyshev_filter_degree(Lar,Las,fstop):
    #KAYNAK:Microstrip Filters for RF Microwave Applications s.41, fstop: normalize frekans, Las: fstoptaki bastirma (dB), Lar: bant ici ripple (dB)
    temp=csqrt(((pow(10.0,0.1*Las)-1)/(pow(10.0,0.1*Lar)-1)))
    return int((arccosh(temp)/arccosh(fstop)))+1

def chebyshev_filter_prototype(n,Lar):
    #n-odd, KAYNAK:Microstrip Filters for RF Microwave Applications s.42, Lar: passband ripple in dB
    beta=-log(tanh((Lar/17.37)))
    gamma=sinh(beta/2/n)
    g=list(range(n+2))
    g[0]=g[n+1]=1.0
    g[1]=2/gamma*sin(0.5*pi/n)
    i=2
    while i<n+1:
        g[i]=1./g[i-1]*4*sin((2*i-1)*pi/2/n)*sin((2*i-3)*pi/2/n)/(gamma**2+sin((i-1)*pi/n)**2)
        i=i+1
    return g

def lp_filter_from_prototype(g,Zo,fc,type=1):
    #n-odd, KAYNAK:Microstrip Filters for RF Microwave Applications s.42
    #type=1, starts with L
    #type=2, starts with C
    n=len(g)-2
    g=g[:1]+[k/2/pi/fc for k in g[1:-1]]+g[-1:]
    sonuc=""
    if type==1:
        for i in range(1,n+1):
            if i%2==1:
                g[i]=g[i]*Zo
                sonuc=sonuc+str(i)+"-Series L- "+str(g[i])+" H\n"
            else:
                g[i]=(g[i]/Zo)
                sonuc=sonuc+str(i)+"-Shunt C-"+str(g[i])+" F\n"
    elif type==2:
        for i in range(1,n+1):
            if i%2==1:
                g[i]=(g[i]/Zo)
                sonuc=sonuc+str(i)+"-Shunt C-"+str(g[i])+" F\n"
            else:
                g[i]=g[i]*Zo
                sonuc=sonuc+str(i)+"-Series L-"+str(g[i])+" H\n"
    return sonuc

def hp_filter_from_prototype(g,Zo,fc,type=1):
    #n-odd, KAYNAK:Microstrip Filters for RF Microwave Applications s.42
    #type=1, starts with L
    #type=2, starts with C
    n=len(g)-2
    g=g[:1]+[1./2./pi/fc/k for k in g[1:-1]]+g[-1:]
    sonuc=""
    if type==1:
        for i in range(1,n+1):
            if i%2==1:
                g[i]=g[i]*Zo
                sonuc=sonuc+str(i)+"-Shunt L-"+str(g[i])+" H\n"
            else:
                g[i]=(g[i]/Zo)
                sonuc=sonuc+str(i)+"-Series C-"+str(g[i])+" F\n"
    elif type==2:
        for i in range(1,n+1):
            if i%2==1:
                g[i]=(g[i]/Zo)
                sonuc=sonuc+str(i)+"-Series C-"+str(g[i])+" F\n"
            else:
                g[i]=g[i]*Zo
                sonuc= sonuc+str(i)+"-Shunt L-"+str(g[i])+" H\n"
    return sonuc

def bp_filter_from_prototype(g,Zo,fc,FBW,type=1):
    #n-odd, KAYNAK:Microstrip Filters for RF Microwave Applications s.42
    #type=1, starts with L
    #type=2, starts with C
    n=len(g)-2
    sonuc=""
    if type==1:
        for i in range(1,n+1):
            if i%2==1:
                sonuc=sonuc+" Ls"+str(i)+" "+str(g[i]*Zo/FBW/2./pi/fc)+"H\t Cs"+str(i)+" "+str(FBW/2./pi/fc/Zo/g[i])+"F\n"
            else:
                sonuc=sonuc+" Lp"+str(i)+" "+str(Zo*FBW/2./pi/fc/g[i])+"H\t Cp"+str(i)+" "+str(1./FBW/2./pi/fc/Zo*g[i])+"F\n"
    elif type==2:
        for i in range(1,n+1):
            if i%2==1:
                sonuc=sonuc+" Lp"+str(i)+" "+str(Zo*FBW/2./pi/fc/g[i])+"H\t Cp"+str(i)+" "+str(1./FBW/2./pi/fc/Zo*g[i])+"F\n"
            else:
                sonuc=sonuc+" Ls"+str(i)+" "+str(g[i]*Zo/FBW/2./pi/fc)+"H\t Cs"+str(i)+" "+str(FBW/2./pi/fc/Zo/g[i])+"F\n"
    return sonuc

def bs_filter_from_prototype(g,Zo,fc,FBW,type=1):
    #n-odd, KAYNAK:Microstrip Filters for RF Microwave Applications s.42
    #type=1, starts with L
    #type=2, starts with C
    n=len(g)-2
    sonuc=""
    if type==1:
        for i in range(1,n+1):
            if i%2==1:
                sonuc=sonuc+" Lp"+str(i)+" "+str(g[i]*Zo*FBW/2./pi/fc)+"H\t Cp"+str(i)+" "+str(1./FBW/2./pi/fc/Zo/g[i])+"F\n"
            else:
                sonuc=sonuc+" Ls"+str(i)+" "+str(Zo/FBW/2./pi/fc/g[i])+"H\t Cs"+str(i)+" "+str(FBW/2./pi/fc/Zo*g[i])+"F\n"
    elif type==2:
        for i in range(1,n+1):
            if i%2==1:
                sonuc=sonuc+" Ls"+str(i)+" "+str(Zo/FBW/2./pi/fc/g[i])+"H\t Cs"+str(i)+" "+str(FBW/2./pi/fc/Zo*g[i])+"F\n"
            else:
                sonuc=sonuc+" Lp"+str(i)+" "+str(g[i]*Zo*FBW/2./pi/fc)+"H\t Cp"+str(i)+" "+str(1./FBW/2./pi/fc/Zo/g[i])+"F\n"
    return sonuc

def chebyshev_stepped_impedance_lp_filter(Zo,maxreturnloss,N,theta):
    #alpha=sin(theta), theta=ue'lerin bant kenarindaki elektriksel uzunlugu
    rip=csqrt((1./(10.**((maxreturnloss/10.0))-1)))
    alpha=sin(theta/180.*pi)
    RL=1
    Z=[]
    if N%2==0:
        print("N should be an odd number.")
        RL=((csqrt(1+rip**2)-rip)/(csqrt(1+rip**2)+rip))
    eta=sinh(1./N*arcsinh((1./rip)))
    A=(1./eta**2.)
    for i in range(1,N+1):
        x=(2*sin((2*i-1)*pi/2./N)/alpha-alpha/4.*(((eta**2+sin(i*pi/N)**2)/sin((2*i+1)*pi/2./N))+((eta**2+sin((i-1)*pi/N)**2)/sin((2*i-3)*pi/2./N))))
        if i%2==0:
            A=1./A/(eta**2+sin((i-1)*pi/N)**2)
            x=x*A*eta
            Z.append(Zo*x)
        else:
            A=1./A/(eta**2+sin((i-1)*pi/N)**2)
            x=x*A/eta
            Z.append((Zo/x))
    return Z

def inductive_post_wg_filter(er, a, b, maxreturnloss,N, d,  x,  f1, f2):
    #kaynak: theory and design of microwave filters s.220
    fcenter=((f1+f2)/2.)
    fc=co/csqrt(er)/2./a
    lg1=co/csqrt(er)/f1/csqrt(1.-((fc/f1))**2.)
    lg2=co/csqrt(er)/f2/csqrt(1.-((fc/f2))**2.)
    lgo=((lg1+lg2)/2.)+1./pi*(lg1*cos(pi/2.*lg2/lg1)+lg2*cos(pi/2.*lg1/lg2))/(sin(pi/2.*lg2/lg1)+sin(pi/2.*lg1/lg2))
    alpha=(1./(lg1/lgo*sin(pi*lgo/lg1)))
    rip=csqrt((1./(10.**((maxreturnloss/10.0))-1)))
    eta=sinh(1./N*arcsinh((1./rip)))
    r=array(list(range(1, N+1)))
    Z=ones(N+2)
    K=ones(N+1)
    Z[1:N+1]=2*alpha/eta*sin((2*r-1)*pi/2./N) -0.25/eta/alpha*(((eta**2+sin(r*pi/N)**2)/sin((2*r+1)*pi/2./N))+((eta**2+sin((r-1)*pi/N)**2)/sin((2*r-3)*pi/2./N)))
    r=array(list(range( N+1)))
    K=(csqrt(eta**2+sin(r*pi/N)**2)/eta)
    for i in range(N+1):
        K[i]=(K[i]/csqrt(Z[i]*Z[i+1]))
    B=(1./K)-K
    L=1./2./pi/fcenter/B
    fi=ones(N)
    lengths=ones(N)
    for i in range(N):
        fi[i]=(pi-0.5*(arctan((2./B[i]))+arctan((2./B[i+1]))))/pi*180.
        lengths[i]=fi[i]/180./csqrt(4.*er*fcenter**2/co**2-(1./a**2))
    #constant diameter synthesis
    sonuc_x=[]
    def integrand(c):
        ind, cap, z=InductivePostInWaveguide(er, a, b, d, c, fcenter)
        return (ind/z)
    Lmin=integrand((a/2.0))
    for temp in L:
        if temp<Lmin:
            sonuc_x.append(0)
            print("Inductance truncated to the minimum achievable")
        else:
            sonuc_x.append(brentq(lambda x:integrand(x)-temp,(d/2.0),(a/2.),   maxiter=200))

    #constant offset synthesis
    sonuc_d=[]
    def integrand(c):
        ind, cap, z=InductivePostInWaveguide(er, a, b, c, x, fcenter)
        return (ind/z)
    for temp in L:
        try:
            sonuc_d.append(brentq(lambda x:integrand(x)-temp,(a/1000.),a,   maxiter=200))
        except:
            print("Error at calculating diameter, L="+str(temp))
    return (L,fi, sonuc_x, sonuc_d, lengths)

def filter_with_j_inverter(g,Zo,fcenter,fbw,caps):
    #order=N ise g'nin boyutu n+2 kapasite sayisi ise n'dir ve biz belirleriz.J-inverter sayisi n+1'dir
    inds=[]
    n=len(g)-2
    Yo=(1./Zo)
    jvalues=list(range(n+1))
    for i in range(1,n):
        jvalues[i]=fbw*2*pi*fcenter*csqrt(caps[i-1]*caps[i]/g[i]/g[i+1])
    jvalues[0]=csqrt(fbw*2*pi*fcenter*caps[0]*Yo/g[0]/g[1])
    jvalues[n]=csqrt(fbw*2*pi*fcenter*caps[n-1]*Yo/g[n]/g[n+1])
    inds=1./(2.*pi*fcenter)**2/array(caps)
    return jvalues,inds,caps

def filter_with_j_inverter2(g,Zo,fcenter,fbw,caps): #first and last inverters are eliminated
    #order=N ise g'nin boyutu n+2 kapasite sayisi ise n'dir ve biz belirleriz.J-inverter sayisi n+1'dir
    inds=[]
    n=len(g)-2
    Yo=(1./Zo)
    jvalues=list(range(n+1))
    for i in range(1,n):
        jvalues[i]=fbw*2*pi*fcenter*csqrt(caps[i-1]*caps[i]/g[i]/g[i+1])
    jvalues[0]=csqrt(fbw*2*pi*fcenter*caps[0]*Yo/g[0]/g[1])
    jvalues[n]=csqrt(fbw*2*pi*fcenter*caps[n-1]*Yo/g[n]/g[n+1])
    inds=1./(2.*pi*fcenter)**2/array(caps)
    return (jvalues[1:-1]/(Zo*Zo*jvalues[0]*jvalues[0])),inds*Zo*Zo*jvalues[0]*jvalues[0],(caps/(Zo*Zo*jvalues[0]*jvalues[0]))


def evanescent_wg_filter3(g,n,Lj,a,a1,b,er, fcenter,fbw,alpha):
    #order=3,
    #n:WG junction transformer ratio (@fcenter)
    #Zj:abs(WG junction impedance)  (@fcenter)
    #alpha: J-coupled 3 order filtrede C2/C1
    #a: width of TE10 waveguide
    #a1: width of evanescent waveguide
    lcutoff=2*a1
    wavelength=co/csqrt(er)/fcenter
    w0=2.0*pi*fcenter
    Xo=abs(zwg_te10(er,a1,b,fcenter))
    Zo=abs(zwg_te10(er,a,b,fcenter))
    #Zo=100.0
    C2=g[0]*g[1]*alpha/fbw/w0/n/n/Zo
    C1=g[0]*g[1]/fbw/w0/n/n/Zo
    L2=1.0/w0/w0/C2
    L1=alpha*L2  #fbw*n*n*Zo/(w0*g0*g1)
    gamma=2.*pi/wavelength*csqrt(((wavelength/lcutoff))**2.0-1.+0.0j)
    J=g[0]*g[1]*csqrt(alpha/g[1]/g[2])/n/n/Zo
    length=(arcsinh(abs(1.0/Xo/J))/gamma)
    L=Xo*tanh(gamma*length)/w0
    L1=L1*L/(L-L1)
    L1=L1*Lj/(Lj-L1/n/n)
    L2=L2*Xo*tanh(gamma*length)/(Xo*tanh(gamma*length)-2*w0*L2)
    return (length,C1,C2,L1,L2,J)

def evanescent_wg_filter4(g,n,Lj,a,a1,b,er, fcenter,fbw,alpha):
    #generalized version of EvanescentWGFilter_3
    #n:WG junction transformer ratio (@fcenter)
    #Zj:abs(WG junction impedance)  (@fcenter)
    #alpha: J-coupled 3 order filtrede C2/C1
    #a: width of TE10 waveguide
    #a1: width of evanescent waveguide
    lcutoff=2*a1
    wavelength=co/csqrt(er)/fcenter
    w0=2.0*pi*fcenter
    Xo=abs(zwg_te10(er,a1,b,fcenter))
    Zo=abs(zwg_te10(er,a,b,fcenter))
    #Zo=100.0
    C2=g[0]*g[1]*alpha/fbw/w0/n/n/Zo
    C1=g[0]*g[1]/fbw/w0/n/n/Zo
    L2=1.0/w0/w0/C2
    L1=alpha*L2  #fbw*n*n*Zo/(w0*g0*g1)
    gamma=2.*pi/wavelength*csqrt(((wavelength/lcutoff))**2.0-1.+0.0j)
    J=g[0]*g[1]*csqrt(alpha/g[1]/g[2])/n/n/Zo
    length=(arcsinh(abs(1.0/Xo/J))/gamma)
    L=Xo*tanh(gamma*length)/w0
    L1=L1*L/(L-L1)
    L1=L1*Lj/(Lj-L1/n/n)
    L2=L2*Xo*tanh(gamma*length)/(Xo*tanh(gamma*length)-2*w0*L2)
    return (length,C1,C2,L1,L2,J)

if __name__ == "__main__":
    #print ButterworthFilterPrototype(5)
    from pylab import *
    from network import *
    a=0.016
    a1=0.008
    b=0.0005
    #b=0.001524
    er=3.38
    c2_c1=2
    n=1.6       #symmetric
    X=0.4e-9    #symmetric

    n=1.6       #symmetric
    X=1.2e-9    #symmetric

    #n=2.1       #asymmetric
    #X=0.16e-9   #asymmetric
    loss=1 #plot loss if 1, plot isolation if 0
    (length,c1,c2,l1, l2, j)=EvanescentWGFilter_3(ButterworthFilterPrototype(3),n,X,a,a1, b,er, 8.5e9,0.25,c2_c1)
    print("length= ",length)
    print("C1= ",c1)
    print("C2= ",c2)
    print("L1= ",l1)
    print("L2= ",l2)
    if l1<0:
        l1=10000
    if l2<0:
        l2=10000
    sonuc=[]
    #length=0.0042
    #l2=0.1145e-9
    #c2=4.847e-12
    l2i=0.042e-9
    r2i=0.26
    l1i=0.16e-9
    r1i=0.98

    for frek in linspace(6e9,10e9,101):
        networks=[]
        networks.append(shunt_z(2.0j*pi*frek*X))
        networks.append(transformer(n))
        if loss:
            networks.append(shunt_z(-1.0j/2./pi/frek/c1))
            networks.append(shunt_z(2.0j*pi*frek*l1))
        else:
            networks.append(shunt_z(2.0j*pi*frek*l1i+r1i))

        networks.append(EWG_ABCD(a1,b,er,length,frek))
        #networks.append(EWG_inv(a1,b,er,length,frek))
        #networks.append(jinv(j))
        #networks.append(shunt_z(-1.0j/j))
        #networks.append(series_z(1.0j/j))
        #networks.append(shunt_z(-1.0j/j))
        if loss:
            networks.append(shunt_z(-1.0j/2./pi/frek/c2))
            networks.append(shunt_z(2.0j*pi*frek*l2))
        else:
            networks.append(shunt_z(2.0j*pi*frek*l2i+r2i))
        networks.append(EWG_ABCD(a1,b,er,length,frek))
        #networks.append(EWG_inv(a1,b,er,length,frek))
        #networks.append(jinv(j))
        #networks.append(shunt_z(-1.0j/j))
        #networks.append(series_z(1.0j/j))
        #networks.append(shunt_z(-1.0j/j))
        if loss:
            networks.append(shunt_z(-1.0j/2./pi/frek/c1))
            networks.append(shunt_z(2.0j*pi*frek*l1))
        else:
            networks.append(shunt_z(2.0j*pi*frek*l1i+r1i))

        networks.append(transformer((1./n)))
        networks.append(shunt_z(2.0j*pi*frek*X))
        abcd=cascade_networks(networks)
        sonuc.append(abcd2s(abcd,zwg_te10(er,a,b,frek))[0,1])
    #sonuc.append(abcd2s(abcd,zwg_te10(er,a,b,8.5e9))[0,1])
    plot(linspace(6e9,10e9,101),20.0*log10(abs(array(sonuc))))
    grid()
    xlabel("Frequency")
    if loss:
        ylabel("Insertion Loss")
    else:
        ylabel("Isolation")
    show()
    print(filter_with_inverter2(ButterworthFilterPrototype(3),16,8.5e9,0.125,[0.1e-12,0.5e-12,0.1e-12]))



#-*-coding:utf-8-*-
"""
Network Parameters
"""
import numpy as np
import sympy as sp
import operator as op
from functools import reduce

def ideal_nport(N):
    """
    S-parameters of ideal N-port junction with equal reference impedances at all ports.
    """
    S = sp.ones(N,N)
    a=sp.Rational(2-N,N)
    b=sp.Rational(2,N)
    S=S*b
    for i in range(N):
        S[i,i]=a
    return S

def ideal_gyrator():
    """
    S-parameters of ideal gyrator.
    """
    return sp.Matrix([[0,-1],[1,0]])

def ideal_coupled_line(Ze, Zo, Te, To, Z0):
    """
    S-parameters of ideal coupled line.
    Te and To in radian
    3----------4
    1----------2
    """
    De=2*Ze*Z0*sp.cos(Te)+(Ze**2+Z0**2)*sp.sin(Te)*sp.I
    Do=2*Zo*Z0*sp.cos(To)+(Zo**2+Z0**2)*sp.sin(To)*sp.I
    Ye=Ze*Z0/De
    Yo=Zo*Z0/Do
    Xe=(Ze**2-Z0**2)*sp.sin(Te)*sp.I/2/De
    Xo=(Zo**2-Z0**2)*sp.sin(To)*sp.I/2/Do
    S = sp.ones(4,4)
    S[0,0]=S[1,1]=S[2,2]=S[3,3]=Xe+Xo
    S[0,1]=S[1,0]=S[2,3]=S[3,2]=Ye+Yo
    S[0,2]=S[2,0]=S[1,3]=S[3,1]=Xe-Xo
    S[0,3]=S[3,0]=S[1,2]=S[2,1]=Ye-Yo
    return S

def ideal_amp(G):
    """
    S-parameters of an ideal amplifier/isolator.
    G is voltage gain, no reflection, infinite isolation.
    """
    return sp.Matrix([[0, 0], [G, 0]])

def ideal_att(G):
    """
    S-parameters of an ideal amplifier/attenuator/isolator.
    G is voltage gain (<1), no reflection.
    """
    return sp.Matrix([[0, G], [G, 0]])

def circulator():
    """
    S-parameters of an ideal circulator (circulation direction 1⇒2⇒3).
    """
    return sp.Matrix([[0, 0, 1], [1, 0, 0], [0, 1, 0]])

def shunt_z(Z):
    """
    ABCD parameters of shunt impedance.
    """
    return sp.Matrix([[1, 0], [1/Z, 1]])

def series_z(Z):
    """
    ABCD parameters of series impedance.
    """
    return sp.Matrix([[1, Z], [0, 1]])

def jinv(J):
    """
    ABCD parameters of J-inverter.
    """
    return sp.Matrix([[0, sp.I/J], [sp.I * J, 0]])

def jinv_lumped(X):
    """
    ABCD parameters of J-inverter produced by 3 inductors in Tee form.
    """
    return CascadeNetworks([shZ( - X), seZ(X), shZ( - X)])

def kinv(K):
    """
    ABCD parameters of k-inverter.
    """
    return sp.Matrix([[0,sp.I * K], [sp.I/K, 0]])

def tline(Zo, theta):
    """
    ABCD parameters of ideal transmission line,  theta in radian.
    """
    return sp.Matrix([[sp.cos(theta), sp.I * Zo * sp.sin(theta)], [sp.I/Zo * sp.sin(theta), sp.cos(theta)]])

def tline_list(Zo, theta):
    """
    ABCD parameters of ideal transmission line,  theta in radian.
    """
    return tuple([sp.cos(theta), sp.I * Zo * sp.sin(theta), sp.I/Zo * sp.sin(theta), sp.cos(theta)])

def transformer(N):
    """
    ABCD parameters of ideal transformer (1:N).
    """
    return sp.Matrix([[1/N, 0], [0, N]])

def t_network(Zs1, Zp, Zs2):
    """
    ABCD parameters of Tee network.
    """
    return sp.Matrix([[1 + Zs1/Zp, Zs1 + Zs2 + Zs1 * Zs2/Zp], [1/Zp, 1 + Zs2/Zp]])

def pi_network(Zp1, Zs, Zp2):
    """
    ABCD parameters of Pi network.
    """
    return sp.Matrix([[1 + Zs/Zp2, Zs], [1/Zp1 + 1/Zp2 + Zs/Zp1/Zp2, 1 + Zs/Zp1]])

def abcd2y(M):
    """
    ABCD parameters to Y-Parameters conversion.
    """
    a, b, c, d = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    return sp.Matrix([[d/b, (b * c - a * d)/b], [ -1/b, a/b]])

def y2abcd(M):
    """
    Y-Parameters to ABCD parameters conversion.
    """
    y11, y12, y21, y22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    return sp.Matrix([[ -y22/y21,  -1/y21], [(y12 * y21 - y11 * y22)/y21,  -y11/y21]])

def t2s(M):
    """
    Transfer scattering parameters to S-Parameters conversion.
    According to definition [b1,a1]=T.[a2,b2].
    Ref: https://en.wikipedia.org/wiki/Scattering_parameters#Scattering_transfer_parameters
    """
    t11, t12, t21, t22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    delta=t11*t22-t12*t21
    return sp.Matrix([[ t12/t22,  delta/t22], [1/t22,  -t21/t22]])

def s2t(M):
    """
    S-Parameters to Transfer scattering parameters conversion.
    According to definition [b1,a1]=T.[a2,b2].
    Ref: https://en.wikipedia.org/wiki/Scattering_parameters#Scattering_transfer_parameters
    """
    s11, s12, s21, s22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    delta=s11*s22-s12*s21
    return sp.Matrix([[ -delta/s21,  s11/s21], [-s22/s21,  1/s21]])

def abcd2z(M):
    """
    ABCD parameters to Z-Parameters conversion
    """
    a, b, c, d = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    return sp.Matrix([[a/c, ( - b * c + a * d)/c], [1/c, d/c]])

def z2abcd(M):
    """
    Z-Parameters to ABCD parameters conversion.
    """
    z11, z12, z21, z22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    return sp.Matrix([[z11/z21, (z11 * z22 - z21 * z12)/z21], [1/z21, z22/z21]])

def abcd2h(M):
    """
    ABCD parameters to H-Parameters conversion.
    """
    a, b, c, d = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    return sp.Matrix([[b/d, ( - b * c + a * d)/d],
                      [-1/d, c/d]])

def h2abcd(M):
    """
    H-Parameters to ABCD parameters conversion.
    """
    h11, h12, h21, h22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    return sp.Matrix([[h12 - h22 * h11 / h21, -h11 / h21],
                      [-h22 / h21, -1 / h21]])

def h2s(M, Zo=50):
    """
    H-Parameters to S parameters conversion. Valid for real Zo value.
    """
    h11, h12, h21, h22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    delta = (h11 + Zo) * (h22 * Zo + 1) - h12 * h21 * Zo
    return sp.Matrix([[((h11 - Zo) * (h22 * Zo + 1) - h12 * h21 * Zo) / delta, 2 * h12 * Zo / delta],
                      [-2 * h12 * Zo / delta, ((h11 + Zo) * (1 - h22 * Zo) + h12 * h21 * Zo) / delta]])

def s2h(M, Zo=50):
    """
    S-Parameters to H parameters conversion. Valid for real Zo value.
    """
    s11, s12, s21, s22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    delta = (1 - s11) * (1 + s22) + s12 * s21
    return sp.Matrix([[Zo * ((1 + s11) * (1 + s22) - s12 * s21) / delta, 2 * s12 / delta],
                      [-2 * s21 / delta, 1 / Zo * ((1 - s11) * (1 - s22) - s12 * s21) / delta]])


def abcd2s(M, Zo=50):
    """
    ABCD parameters to S-Parameters conversion.
    Valid for real Zo value.
    """
    # Zo=sp.Symbol("Zo")
    a, b, c, d = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    s11 = ((a + b/Zo - c * Zo - d)/(a + b/Zo + c * Zo + d))
    s12 = 2 * (a * d - b * c)/(a + b/Zo + c * Zo + d)
    s21 = (2/(a + b/Zo + c * Zo + d))
    s22 = (( - a + b/Zo - c * Zo + d)/(a + b/Zo + c * Zo + d))
    return sp.Matrix([[s11, s12], [s21, s22]])

def abcd2s_list(M, Zo=50):
    """
    ABCD parameters to S-Parameters conversion
    Valid for real Zo value
    """
    a, b, c, d = M[0], M[1], M[2], M[3]
    s11 = ((a + b/Zo - c * Zo - d)/(a + b/Zo + c * Zo + d))
    s12 = 2. * (a * d - b * c)/(a + b/Zo + c * Zo + d)
    s21 = (2./(a + b/Zo + c * Zo + d))
    s22 = (( - a + b/Zo - c * Zo + d)/(a + b/Zo + c * Zo + d))
    return [s11, s12, s21, s22]

def s2abcd(M, Z=( 50, 50 )):
    """
    S-Parameters to ABCD parameters conversion.
    Valid for real Z values.
    Z: reference impedance list [Z1, Z2]
    """
    Zo1, Zo2 = tuple(Z)
    s11, s12, s21, s22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    a = ((Zo1 + s11 * Zo1) * (1 - s22) + s12 * s21 * Zo1)/(2. * s21 * sp.sqrt(Zo1 * Zo2))
    b = ((Zo1 + s11 * Zo1) * (Zo2 + s22 * Zo2) - s12 * s21 * Zo1 * Zo2)/(2. * s21 * sp.sqrt(Zo1 * Zo2))
    c = ((1 - s11) * (1 - s22) - s12 * s21)/(2. * s21 * sp.sqrt(Zo1 * Zo2))
    d = ((1 - s11) * (Zo2 + s22 * Zo2) + s12 * s21 * Zo2)/(2. * s21 * sp.sqrt(Zo1 * Zo2))
    return sp.Matrix([[a, b], [c, d]])

def abcd2t(M, Zo=50):
    """
    ABCD parameters to T-Parameters conversion

    ABCD: [V1 I1]=ABCD*[V2 -I2]
    Pseudo-Wave or Power-Wave? Don't use.
    """
    X = abcd2s(M, Zo)
    return s2t(M)

def abcd_change_ports(M):
    """
    Switching ports of ABCD parameters
    """
    M = M.I
    sp.Matrix([[M[0, 0],  - M[0, 1]], [ - M[1, 0], M[1, 1]]])

def t2abcd(M, Z=( 50,50 )):
    """
    T-parameters to ABCD parameters conversion
    """
    X = t2s(M)
    return s2abcd(X, Z)

def snp2smp(SM,ports):
    """
    This method changes the port numbering of the network
    port j of new network corresponds to ports[j] in old network
    if the length of "ports" argument is lower than number of ports, remaining ports are terminated with current reference impedances and number of ports are reduced.
    """
    ports = [x-1 for x in ports]
    ixgrid = ports, ports
    return SM.extract(*ixgrid)

def cascade_networks(networks):
    """
    Cascading 2-port Networks,  input and output is ABCD matrices of networks
    """
    return reduce(op.mul, networks)

def parallel_networks(networks):
    """
    Paralleling 2-port Networks,  input and output  is ABCD matrices of networks
    """
    ymatrices = [abcd2y(M) for M in networks]
    return y2abcd(reduce(op.add, ymatrices))

def series_networks(networks):
    """
    Series Connection of Networks (reference pins of 1. network is connected to alive pins of 2. network),  input and output  is ABCD matrices of networks
    """
    zmatrices = [abcd2z(M) for M in networks]
    return z2abcd(reduce(op.add, zmatrices))

def s_normalize_pseudo(S, Zold, Znew):
    """
    Zold,  Znew port_sayisi uzunlugunda dizilerdir
    Pseudo-Wave icin
    """
    ps = len(Zold)
    A = sp.zeros(ps, ps)
    gamma = sp.zeros(ps, ps)
    imp_yeni = []
    for i in range(ps):
        imp_yeni.append(Znew[i])
        z = Zold[i]
        ri = ((imp_yeni[i] - z)/(imp_yeni[i] + z))
        gamma[i, i] = ri
        A[i, i] = csqrt(z.real/imp_yeni[i].real) * abs(imp_yeni[i]/z) * 2 * z/(imp_yeni[i] + z)
    return A.I * (S - gamma) * (sp.eye(ps) - gamma * S).I * A

def s_normalize_power(S, Zold, Znew):
    """
    Zold,  Znew are arrays with the same size as number of ports.
    Power-Wave icin
    Reference: Article, “Multiport conversions between S, Z, Y, h, ABCD, and T parameters”
    """
    ps = len(Zold)
    A = sp.zeros(ps, ps)
    gamma = sp.zeros(ps, ps)
    imp_yeni = []
    for i in range(ps):
        imp_yeni.append(Znew[i])
        z = Zold[i]
        gamma[i, i] = ((imp_yeni[i] - z)/(imp_yeni[i] + z.C))
        A[i, i] = 2*csqrt(z.real)*csqrt(imp_yeni[i].real) /(imp_yeni[i].C + z)
    return A.I * (S - gamma.C) * (sp.eye(ps) - gamma * S).I * A.C

def s_phase_deembed(S, phase):
    """
    S-parameter deembedding
    S is sympy.Matrix NxN
    phase, deembedding phase for each port in radian. Positive phase is deembedding into the circuit
    """
    PhaseMatrix=sp.exp(1j*sp.diag(*phase))
    return PhaseMatrix*S*PhaseMatrix

def connect_2_ports(Smatrix,k,m):
    """ Port-m is connected to port-k and both ports are removed.
    Reference: QUCS technical.pdf, S-parameters in CAE programs, p.29.
    """
    k,m=min(k,m),max(k,m)
    k, m = k-1, m-1
    ps=Smatrix.shape[0]
    Sm=sp.zeros(ps-2,ps-2) # output matrix
    S = Smatrix
    for i in range(ps-2):
        ii=i+(i>=k)+(i>=(m-1))
        for j in range(ps-2):
            jj=j+(j>=k)+(j>=(m-1))
            # index = (ps-2)*(i-1)+(j-1)
            temp = S[k,jj]*S[ii,m]*(1-S[m,k])+S[m,jj]*S[ii,k]*(1-S[k,m])+S[k,jj]*S[m,m]*S[ii,k]+S[m,jj]*S[k,k]*S[ii,m]
            Sm[i,j] = S[ii,jj] + temp/((1-S[m,k])*(1-S[k,m])-S[k,k]*S[m,m])
    return Sm

def connect_network_1_conn(Smatrix,EX,k,m):
    """ Port-m of EX circuit is connected to port-k of this circuit.
    Remaining ports of EX are added to the port list of this circuit in order.
    Reference: QUCS technical.pdf, S-parameters in CAE programs, p.29
    """
    S = Smatrix
    k, m = k-1, m-1
    ps1=S.shape[0]
    ps2=EX.shape[0]
    ps=ps1+ps2-2
    Sm=sp.ones(ps,ps)
    for i in range(ps1-1):
        ii=i+(i>(k-1))
        for j in range(ps1-1):
            jj=j+(j>(k-1))
            # index = (i-1)*ps+(j-1)
            Sm[i,j] = S[ii,jj]+S[k,jj]*EX[m,m]*S[ii,k]/(1-S[k,k]*EX[m,m])
    for i in range(ps2-1):
        ii=i+(i>(m-1))
        for j in range(ps1-1):
            jj=j+(j>(k-1))
            # index = (i+ps1-1-1)*ps+(j-1)
            Sm[i+ps1-1,j] = S[k,jj] * EX[ii,m] / (1 - S[k,k] * EX[m,m])
    for i in range(ps1-1):
        ii=i+(i>(k-1))
        for j in range(ps2-1):
            jj=j+(j>(m-1))
            # index = (i-1)*ps+(j+ps1-1-1)
            Sm[i,j+ps1-1] = EX[m,jj] * S[ii,k] / (1 - EX[m,m] * S[k,k])
    for i in range(ps2-1):
        ii=i+(i>(m-1))
        for j in range(ps2-1):
            jj=j+ (j>(m-1))
            # index = (i+ps1-1-1)*ps+(j+ps1-1-1)
            Sm[i+ps1-1,j+ps1-1] = EX[ii,jj]+EX[m,jj]*S[k,k]*EX[ii,m]/(1-EX[m,m]*S[k,k])
    return Sm

def connect_network_1_conn_retain(Smatrix,EX,k,m):
    ideal3port = ideal_nport(3)
    EX = connect_network_1_conn(EX,ideal3port,m,1)
    psex = EX.shape[0]
    sonuc = connect_network_1_conn(Smatrix,EX,k,psex)
    return sonuc

def connect_2_ports_list(Smatrix,conns):
    """ Short circuit ports together one-to-one. Short circuited ports are removed.
    Ports that will be connected are given as tuples in list conn
    i.e. conn=[(p1,p2),(p3,p4),..]
    The order of remaining ports is kept.
    Reference: QUCS technical.pdf, S-parameters in CAE programs, p.29
    """
    for i in range(len(conns)):
        k,m = conns[i]
        Smatrix = connect_2_ports(Smatrix, k,m)
        for j in range(i+1,len(conns)):
            conns[j][0]=conns[j][0]-(conns[j][0]>k)-(conns[j][0]>m)
            conns[j][1]=conns[j][1]-(conns[j][1]>k)-(conns[j][1]>m)
    return Smatrix

def connect_2_ports_retain(Smatrix,k,m):
    """ Port-m and Port-k are joined to a single port.
    New port becomes the last port of the circuit.
    Reference: QUCS technical.pdf, S-parameters in CAE programs, p.29
    """
    ideal3port = idealNport(3)
    ps = Smatrix.shape[0]
    k,m = min(k,m),max(k,m)
    Sm = connect_network_1_conn(Smatrix,ideal3port,m,1)
    Sm = connect_2_ports(Sm,k,ps)
    return Sm

if __name__=="__main__":
    R1, R2 = sp.symbols("R1 R2")
    # s1 = abcd2s(shunt_z(R1)*series_z(R2)*shunt_z(R1) )
    # sonuc = s1.subs([(R1,100),(R2,50)]).evalf()
    # sp.pprint(s1)
    # sp.pprint(sonuc)
    s1=abcd2s(shunt_z(R1))
    s2=abcd2s(series_z(R2))
    # ss = connect_network_1_conn(s1,s2,2,1)
    # ss = connect_network_1_conn(connect_network_1_conn(s1,s2,2,1),s1,2,1)
    # sonuc = ss.subs([(R1,100),(R2,50)]).evalf()
    # sp.pprint(sp.simplify(s2))
    # sp.pprint(sp.simplify(s2abcd(ss)))
    # sp.pprint(sonuc)
    # sp.pprint(sp.ones(3,4))
    # sp.pprint(sp.diag(*np.array([1,2,3])))
    tri = idealNport(3)
    sp.pprint(sp.simplify(s2))
    # s3 = connect_network_1_conn(tri,s2,2,1)
    # sp.pprint(sp.simplify(s3))
    # s4 = connect_network_1_conn(s3,tri,3,1)
    # s5 = connect_network_1_conn(s4,s2,2,1)
    # s6 = connect_2_ports(s5,2,4)
    # sp.pprint(sp.simplify(s6))

    s4 = connect_network_1_conn_retain(s2,s2,2,2)
    sp.pprint(sp.simplify(s4))
    s5 = connect_2_ports_retain(s4,2,1)
    sp.pprint(sp.simplify(s5))



from mwtoolbox.rfnetwork import *
import mwtoolbox.transmission_lines as tline

w = 25*25.4e-6
h1 = 10*25.4e-6
h2 = 15*25.4e-6
er = 3
t = 1*25.4e-6

z1 = tline.z_qs_thick_embedded_microstrip_1(w, h1, h2, er, t)

z2 = tline.z_qs_thick_embedded_microstrip(w, h1, h2, er, t)

print(f"{z1=}")
print(f"{z2=}")
